"""
Tests for SymmetricKey.
"""
from secrets import token_bytes

import pytest

from cwt.exceptions import DecodeError, EncodeError, VerifyError
from cwt.key_types.symmetric import (
    AESCCMKey,
    AESGCMKey,
    ChaCha20Key,
    HMACKey,
    SymmetricKey,
)


class TestSymmetricKey:
    """
    Tests for SymmetricKey.
    """

    def test_symmetric_key_constructor_with_hmac_256_256(self):
        key = SymmetricKey(
            {
                1: 4,
                -1: b"mysecret",
                3: 5,  # HMAC 256/256
            }
        )
        assert key.key == b"mysecret"
        assert key.kty == 4
        assert key.kid is None
        assert key.alg == 5
        assert key.key_ops is None
        assert key.base_iv is None

    @pytest.mark.parametrize(
        "invalid, msg",
        [
            (
                {1: 2},
                "kty(1) should be Symmetric(4).",
            ),
            (
                {1: b"invalid"},
                "kty(1) should be int or str(tstr).",
            ),
            (
                {1: {}},
                "kty(1) should be int or str(tstr).",
            ),
            (
                {1: []},
                "kty(1) should be int or str(tstr).",
            ),
            (
                {1: 4, -1: 123},
                "k(-1) should be bytes(bstr).",
            ),
            (
                {1: 4, -1: {}},
                "k(-1) should be bytes(bstr).",
            ),
            (
                {1: 4, -1: []},
                "k(-1) should be bytes(bstr).",
            ),
            (
                {1: 4, -1: b"mysecret"},
                "alg(3) not found.",
            ),
        ],
    )
    def test_symmetric_key_constructor_with_invalid_args(self, invalid, msg):
        with pytest.raises(ValueError) as err:
            SymmetricKey(invalid)
            pytest.fail("SymmetricKey should fail.")
        assert msg in str(err.value)


class TestHMACKey:
    """
    Tests for HMACKey.
    """

    def test_hmac_key_constructor_with_hmac_256_256(self):
        key = HMACKey(
            {
                1: 4,
                -1: b"mysecret",
                3: 5,  # HMAC 256/256
            }
        )
        assert key.kty == 4
        assert key.kid is None
        assert key.alg == 5
        assert len(key.key_ops) == 2
        assert 9 in key.key_ops
        assert 10 in key.key_ops
        assert key.base_iv is None
        try:
            sig = key.sign(b"Hello world!")
            key.verify(b"Hello world!", sig)
        except Exception:
            pytest.fail("sign/verify should not fail.")

    def test_hmac_key_constructor_with_hmac_256_256_without_key(self):
        key = HMACKey(
            {
                1: 4,
                3: 5,  # HMAC 256/256
            }
        )
        assert key.kty == 4
        assert key.kid is None
        assert key.alg == 5
        assert len(key.key_ops) == 2
        assert 9 in key.key_ops
        assert 10 in key.key_ops
        assert key.base_iv is None
        try:
            sig = key.sign(b"Hello world!")
            key.verify(b"Hello world!", sig)
        except Exception:
            pytest.fail("sign/verify should not fail.")

    @pytest.mark.parametrize(
        "invalid, msg",
        [
            (
                {1: 2},
                "kty(1) should be Symmetric(4).",
            ),
            (
                {1: b"invalid"},
                "kty(1) should be int or str(tstr).",
            ),
            (
                {1: {}},
                "kty(1) should be int or str(tstr).",
            ),
            (
                {1: []},
                "kty(1) should be int or str(tstr).",
            ),
            (
                {1: 4, -1: 123},
                "k(-1) should be bytes(bstr).",
            ),
            (
                {1: 4, -1: {}},
                "k(-1) should be bytes(bstr).",
            ),
            (
                {1: 4, -1: []},
                "k(-1) should be bytes(bstr).",
            ),
            (
                {1: 4, -1: b"mysecret"},
                "alg(3) not found.",
            ),
            (
                {1: 4, -1: b"mysecret", 3: 3},
                "Unsupported or unknown alg(3) for HMAC.",
            ),
            (
                {1: 4, -1: b"mysecret", 3: 8},
                "Unsupported or unknown alg(8) for HMAC.",
            ),
            (
                {1: 4, -1: b"mysecret", 3: 4, 4: [1, 2, 3]},
                "Unknown or not permissible key_ops(4) for MACAuthenticationKey: 1.",
            ),
            (
                {1: 4, -1: b"mysecret", 3: 4, 4: [9, 10, 11]},
                "Unknown or not permissible key_ops(4) for MACAuthenticationKey: 11.",
            ),
        ],
    )
    def test_hmac_key_constructor_with_invalid_args(self, invalid, msg):
        with pytest.raises(ValueError) as err:
            HMACKey(invalid)
            pytest.fail("HMACKey should fail.")
        assert msg in str(err.value)

    def test_hmac_key_sign_with_invalid_args(self):
        key = HMACKey(
            {
                1: 4,
                -1: b"mysecret",
                3: 5,  # HMAC 256/256
            }
        )
        with pytest.raises(EncodeError) as err:
            key.sign(123)
            pytest.fail("sign should fail.")
        assert "Failed to sign." in str(err.value)

    def test_hmac_key_verify_with_invalid_signature(self):
        key = HMACKey(
            {
                1: 4,
                -1: b"mysecret",
                3: 5,  # HMAC 256/256
            }
        )
        assert key.kty == 4
        assert key.kid is None
        assert key.alg == 5
        assert len(key.key_ops) == 2
        assert 9 in key.key_ops
        assert 10 in key.key_ops
        assert key.base_iv is None
        sig = key.sign(b"Hello world!")
        with pytest.raises(VerifyError) as err:
            key.verify(b"Hello world!", sig + b"xxx")
            pytest.fail("verify should fail.")
        assert "Failed to compare digest." in str(err.value)


class TestAESCCMKey:
    """
    Tests for AESCCMKey.
    """

    def test_aesccm_key_constructor_with_aes_ccm_16_64_128(self):
        key = AESCCMKey(
            {
                1: 4,
                -1: token_bytes(16),
                3: 10,  # AES-CCM-16-64-128
            }
        )
        assert key.kty == 4
        assert key.kid is None
        assert key.alg == 10
        assert len(key.key_ops) == 4
        assert 3 in key.key_ops
        assert 4 in key.key_ops
        assert 5 in key.key_ops
        assert 6 in key.key_ops
        assert key.base_iv is None
        nonce = token_bytes(13)
        try:
            encrypted = key.encrypt(b"Hello world!", nonce=nonce)
            assert key.decrypt(encrypted, nonce) == b"Hello world!"
        except Exception:
            pytest.fail("sign/verify should not fail.")

    @pytest.mark.parametrize(
        "key_args, nonce",
        [
            (
                {1: 4, 3: 10},
                token_bytes(13),
            ),
            (
                {1: 4, 3: 11},
                token_bytes(13),
            ),
            (
                {1: 4, 3: 12},
                token_bytes(7),
            ),
            (
                {1: 4, 3: 13},
                token_bytes(7),
            ),
            (
                {1: 4, 3: 30},
                token_bytes(13),
            ),
            (
                {1: 4, 3: 31},
                token_bytes(13),
            ),
            (
                {1: 4, 3: 32},
                token_bytes(7),
            ),
            (
                {1: 4, 3: 33},
                token_bytes(7),
            ),
        ],
    )
    def test_aesccm_key_constructor_with_aes_ccm_without_key(self, key_args, nonce):
        key = AESCCMKey(key_args)
        assert key.kty == 4
        assert key.kid is None
        assert len(key.key_ops) == 4
        assert 3 in key.key_ops
        assert 4 in key.key_ops
        assert 5 in key.key_ops
        assert 6 in key.key_ops
        assert key.base_iv is None
        try:
            encrypted = key.encrypt(b"Hello world!", nonce=nonce)
            assert key.decrypt(encrypted, nonce) == b"Hello world!"
        except Exception:
            pytest.fail("sign/verify should not fail.")

    @pytest.mark.parametrize(
        "invalid, msg",
        [
            (
                {1: 2},
                "kty(1) should be Symmetric(4).",
            ),
            (
                {1: b"invalid"},
                "kty(1) should be int or str(tstr).",
            ),
            (
                {1: {}},
                "kty(1) should be int or str(tstr).",
            ),
            (
                {1: []},
                "kty(1) should be int or str(tstr).",
            ),
            (
                {1: 4, -1: 123},
                "k(-1) should be bytes(bstr).",
            ),
            (
                {1: 4, -1: {}},
                "k(-1) should be bytes(bstr).",
            ),
            (
                {1: 4, -1: []},
                "k(-1) should be bytes(bstr).",
            ),
            (
                {1: 4, -1: b"mysecret"},
                "alg(3) not found.",
            ),
            (
                {1: 4, -1: b"mysecret", 3: 9},
                "Unsupported or unknown alg(9) for AES CCM.",
            ),
            (
                {1: 4, -1: b"mysecret", 3: 34},
                "Unsupported or unknown alg(34) for AES CCM.",
            ),
            (
                {1: 4, -1: b"mysecret", 3: 10},
                "The length of AES-CCM-16-64-128 key should be 16 bytes.",
            ),
            (
                {1: 4, -1: b"mysecret", 3: 11},
                "The length of AES-CCM-16-64-256 key should be 32 bytes.",
            ),
            (
                {1: 4, -1: b"mysecret", 3: 12},
                "The length of AES-CCM-64-64-128 key should be 16 bytes.",
            ),
            (
                {1: 4, -1: b"mysecret", 3: 13},
                "The length of AES-CCM-64-64-256 key should be 32 bytes.",
            ),
            (
                {1: 4, -1: b"mysecret", 3: 30},
                "The length of AES-CCM-16-128-128 key should be 16 bytes.",
            ),
            (
                {1: 4, -1: b"mysecret", 3: 31},
                "The length of AES-CCM-16-128-256 key should be 32 bytes.",
            ),
            (
                {1: 4, -1: b"mysecret", 3: 32},
                "The length of AES-CCM-64-128-128 key should be 16 bytes.",
            ),
            (
                {1: 4, -1: b"mysecret", 3: 33},
                "The length of AES-CCM-64-128-256 key should be 32 bytes.",
            ),
            (
                {1: 4, 3: 10, 4: [1, 2]},
                "Unknown or not permissible key_ops(4) for ContentEncryptionKey: 1.",
            ),
            (
                {1: 4, 3: 10, 4: [3, 4, 11]},
                "Unknown or not permissible key_ops(4) for ContentEncryptionKey: 11.",
            ),
            (
                {1: 4, 3: 10, 4: [5, 6, 11]},
                "Unknown or not permissible key_ops(4) for ContentEncryptionKey: 11.",
            ),
        ],
    )
    def test_aesccm_key_constructor_with_invalid_args(self, invalid, msg):
        with pytest.raises(ValueError) as err:
            AESCCMKey(invalid)
            pytest.fail("AESCCMKey should fail.")
        assert msg in str(err.value)

    def test_aesgcm_key_encrypt_without_msg(self):
        key = AESCCMKey(
            {
                1: 4,
                -1: token_bytes(16),
                3: 10,  # AES-CCM-16-64-128
            }
        )
        nonce = token_bytes(13)
        with pytest.raises(EncodeError) as err:
            key.encrypt(None, nonce=nonce)
        assert "Failed to encrypt." in str(err.value)

    def test_aesccm_key_decrypt_with_invalid_nonce(self):
        key = AESCCMKey(
            {
                1: 4,
                -1: token_bytes(16),
                3: 10,  # AES-CCM-16-64-128
            }
        )
        assert key.kty == 4
        assert key.kid is None
        assert key.alg == 10
        assert len(key.key_ops) == 4
        assert 3 in key.key_ops
        assert 4 in key.key_ops
        assert 5 in key.key_ops
        assert 6 in key.key_ops
        assert key.base_iv is None
        nonce = token_bytes(13)
        encrypted = key.encrypt(b"Hello world!", nonce=nonce)
        with pytest.raises(DecodeError) as err:
            key.decrypt(encrypted, nonce=token_bytes(13))
        assert "Failed to decrypt." in str(err.value)

    def test_aesccm_key_decrypt_with_invalid_length_nonce(self):
        key = AESCCMKey(
            {
                1: 4,
                -1: token_bytes(16),
                3: 10,  # AES-CCM-16-64-128
            }
        )
        assert key.kty == 4
        assert key.kid is None
        assert key.alg == 10
        assert len(key.key_ops) == 4
        assert 3 in key.key_ops
        assert 4 in key.key_ops
        assert 5 in key.key_ops
        assert 6 in key.key_ops
        assert key.base_iv is None
        nonce = token_bytes(13)
        encrypted = key.encrypt(b"Hello world!", nonce=nonce)
        with pytest.raises(ValueError) as err:
            key.decrypt(encrypted, nonce=token_bytes(7))
        assert "The length of nonce should be 13 bytes." in str(err.value)


class TestAESGCMKey:
    """
    Tests for AESGCMKey.
    """

    def test_aesgcm_key_constructor_with_aes_gcm_a128gcm(self):
        key = AESGCMKey(
            {
                1: 4,
                -1: token_bytes(16),
                3: 1,  # A128GCM
            }
        )
        assert key.kty == 4
        assert key.kid is None
        assert key.alg == 1
        assert len(key.key_ops) == 4
        assert 3 in key.key_ops
        assert 4 in key.key_ops
        assert 5 in key.key_ops
        assert 6 in key.key_ops
        assert key.base_iv is None
        nonce = token_bytes(12)
        try:
            encrypted = key.encrypt(b"Hello world!", nonce=nonce)
            assert key.decrypt(encrypted, nonce) == b"Hello world!"
        except Exception:
            pytest.fail("sign/verify should not fail.")

    @pytest.mark.parametrize(
        "key_args",
        [
            {1: 4, 3: 1},
            {1: 4, 3: 2},
            {1: 4, 3: 3},
        ],
    )
    def test_aesgcm_key_constructor_with_aes_ccm_without_key(self, key_args):
        key = AESGCMKey(key_args)
        assert key.kty == 4
        assert key.kid is None
        assert len(key.key_ops) == 4
        assert 3 in key.key_ops
        assert 4 in key.key_ops
        assert 5 in key.key_ops
        assert 6 in key.key_ops
        assert key.base_iv is None
        nonce = token_bytes(12)
        try:
            encrypted = key.encrypt(b"Hello world!", nonce=nonce)
            assert key.decrypt(encrypted, nonce) == b"Hello world!"
        except Exception:
            pytest.fail("sign/verify should not fail.")

    @pytest.mark.parametrize(
        "invalid, msg",
        [
            (
                {1: 4, -1: b"mysecret", 3: 4},
                "Unsupported or unknown alg(3) for AES GCM: 4",
            ),
            (
                {1: 4, -1: b"mysecret", 3: 1},
                "The length of A128GCM key should be 16 bytes.",
            ),
            (
                {1: 4, -1: b"mysecret", 3: 2},
                "The length of A192GCM key should be 24 bytes.",
            ),
            (
                {1: 4, -1: b"mysecret", 3: 3},
                "The length of A256GCM key should be 32 bytes.",
            ),
            (
                {1: 4, 3: 1, 4: [1, 2]},
                "Unknown or not permissible key_ops(4) for ContentEncryptionKey: 1.",
            ),
            (
                {1: 4, 3: 1, 4: [3, 4, 11]},
                "Unknown or not permissible key_ops(4) for ContentEncryptionKey: 11.",
            ),
            (
                {1: 4, 3: 1, 4: [5, 6, 11]},
                "Unknown or not permissible key_ops(4) for ContentEncryptionKey: 11.",
            ),
        ],
    )
    def test_aesgcm_key_constructor_with_invalid_args(self, invalid, msg):
        with pytest.raises(ValueError) as err:
            AESGCMKey(invalid)
            pytest.fail("AESGCMKey should fail.")
        assert msg in str(err.value)

    def test_aesgcm_key_encrypt_with_empty_nonce(self):
        key = AESGCMKey(
            {
                1: 4,
                -1: token_bytes(16),
                3: 1,  # A128GCM
            }
        )
        with pytest.raises(EncodeError) as err:
            key.encrypt(b"Hello world!", nonce=b"")
        assert "Failed to encrypt." in str(err.value)

    def test_aesgcm_key_decrypt_with_invalid_nonce(self):
        key = AESGCMKey(
            {
                1: 4,
                -1: token_bytes(16),
                3: 1,  # A128GCM
            }
        )
        assert key.kty == 4
        assert key.kid is None
        assert key.alg == 1
        assert len(key.key_ops) == 4
        assert 3 in key.key_ops
        assert 4 in key.key_ops
        assert 5 in key.key_ops
        assert 6 in key.key_ops
        assert key.base_iv is None
        nonce = token_bytes(12)
        encrypted = key.encrypt(b"Hello world!", nonce=nonce)
        with pytest.raises(DecodeError) as err:
            key.decrypt(encrypted, nonce=token_bytes(12))
        assert "Failed to decrypt." in str(err.value)


class TestChaCha20Key:
    """
    Tests for ChaCha20Key.
    """

    def test_chacha20_key_constructor(self):
        key = ChaCha20Key(
            {
                1: 4,
                -1: token_bytes(32),
                3: 24,  # ChaCha20/Poly1305
            }
        )
        assert key.kty == 4
        assert key.kid is None
        assert key.alg == 24
        assert len(key.key_ops) == 4
        assert 3 in key.key_ops
        assert 4 in key.key_ops
        assert 5 in key.key_ops
        assert 6 in key.key_ops
        assert key.base_iv is None
        nonce = token_bytes(12)
        try:
            encrypted = key.encrypt(b"Hello world!", nonce=nonce)
            assert key.decrypt(encrypted, nonce) == b"Hello world!"
        except Exception:
            pytest.fail("sign/verify should not fail.")

    @pytest.mark.parametrize(
        "key_args",
        [
            {1: 4, 3: 24},
        ],
    )
    def test_chacha20_key_constructor_without_key(self, key_args):
        key = ChaCha20Key(key_args)
        assert key.kty == 4
        assert key.kid is None
        assert len(key.key_ops) == 4
        assert 3 in key.key_ops
        assert 4 in key.key_ops
        assert 5 in key.key_ops
        assert 6 in key.key_ops
        assert key.base_iv is None
        nonce = token_bytes(12)
        try:
            encrypted = key.encrypt(b"Hello world!", nonce=nonce)
            assert key.decrypt(encrypted, nonce) == b"Hello world!"
        except Exception:
            pytest.fail("sign/verify should not fail.")

    @pytest.mark.parametrize(
        "invalid, msg",
        [
            (
                {1: 4, -1: b"mysecret", 3: 0},
                "Unsupported or unknown alg(3) for ChaCha20: 0",
            ),
            (
                {1: 4, -1: b"mysecret", 3: 24},
                "The length of ChaCha20/Poly1305 key should be 32 bytes.",
            ),
            (
                {1: 4, 3: 24, 4: [1, 2]},
                "Unknown or not permissible key_ops(4) for ContentEncryptionKey: 1.",
            ),
            (
                {1: 4, 3: 24, 4: [3, 4, 11]},
                "Unknown or not permissible key_ops(4) for ContentEncryptionKey: 11.",
            ),
            (
                {1: 4, 3: 24, 4: [5, 6, 11]},
                "Unknown or not permissible key_ops(4) for ContentEncryptionKey: 11.",
            ),
        ],
    )
    def test_chacha20_key_constructor_with_invalid_args(self, invalid, msg):
        with pytest.raises(ValueError) as err:
            ChaCha20Key(invalid)
            pytest.fail("ChaCha20Key should fail.")
        assert msg in str(err.value)

    def test_chacha20_key_encrypt_with_empty_nonce(self):
        key = ChaCha20Key(
            {
                1: 4,
                -1: token_bytes(32),
                3: 24,  # ChaCha20/Poly1305
            }
        )
        with pytest.raises(EncodeError) as err:
            key.encrypt(b"Hello world!", nonce=b"")
        assert "Failed to encrypt." in str(err.value)

    def test_chacha20_key_decrypt_with_different_nonce(self):
        key = ChaCha20Key(
            {
                1: 4,
                -1: token_bytes(32),
                3: 24,  # ChaCha20/Poly1305
            }
        )
        nonce = token_bytes(12)
        encrypted = key.encrypt(b"Hello world!", nonce=nonce)
        with pytest.raises(DecodeError) as err:
            key.decrypt(encrypted, nonce=token_bytes(12))
            pytest.fail("decrypt should fail.")
        assert "Failed to decrypt." in str(err.value)

    def test_chacha20_key_decrypt_with_different_key(self):
        key = ChaCha20Key(
            {
                1: 4,
                -1: token_bytes(32),
                3: 24,  # ChaCha20/Poly1305
            }
        )
        key2 = ChaCha20Key(
            {
                1: 4,
                -1: token_bytes(32),
                3: 24,  # ChaCha20/Poly1305
            }
        )
        nonce = token_bytes(12)
        encrypted = key.encrypt(b"Hello world!", nonce=nonce)
        with pytest.raises(DecodeError) as err:
            key2.decrypt(encrypted, nonce=nonce)
            pytest.fail("decrypt should fail.")
        assert "Failed to decrypt." in str(err.value)

    def test_chacha20_key_decrypt_with_invalid_nonce(self):
        key = ChaCha20Key(
            {
                1: 4,
                -1: token_bytes(32),
                3: 24,  # ChaCha20/Poly1305
            }
        )
        nonce = token_bytes(12)
        encrypted = key.encrypt(b"Hello world!", nonce=nonce)
        with pytest.raises(DecodeError) as err:
            key.decrypt(encrypted, nonce=token_bytes(8))
            pytest.fail("decrypt should fail.")
        assert "Failed to decrypt." in str(err.value)



class CWTError(Exception):
    """
    Base class for all exceptions.
    """

    pass


class VerifyError(CWTError):
    """
    An Exception occurred when a verification process failed.
    """

    pass


class EncodeError(CWTError):
    """
    An Exception occurred when a CWT/COSE encoding process failed.
    """

    pass


class DecodeError(CWTError):
    """
    An Exception occurred when a CWT/COSE decoding process failed.
    """

    pass



# pylint: disable=R0201, R0904, W0621
# R0201: Method could be a function
# R0904: Too many public methods
# W0621: Redefined outer name

"""
Tests for ClaimsBuilder.
"""
import pytest

from cwt import ClaimsBuilder


@pytest.fixture(scope="session", autouse=True)
def ctx():
    return ClaimsBuilder()


class TestClaimsBuilder:
    """
    Tests for ClaimsBuilder.
    """

    def test_claims_builder_constructor(self):
        c = ClaimsBuilder()
        assert isinstance(c, ClaimsBuilder)

    @pytest.mark.parametrize(
        "json, expected",
        [
            ({"iss": "coap://as.example.com"}, {1: "coap://as.example.com"}),
            ({"sub": "erikw"}, {2: "erikw"}),
            ({"aud": "coap://light.example.com"}, {3: "coap://light.example.com"}),
            ({"exp": 1444064944}, {4: 1444064944}),
            ({"nbf": 1443944944}, {5: 1443944944}),
            ({"iat": 1443944944}, {6: 1443944944}),
            ({"cti": "123"}, {7: b"123"}),
            ({}, {}),
        ],
    )
    def test_claims_builder_from_json(self, ctx, json, expected):
        claims = ctx.from_json(json).to_dict()
        for k, v in claims.items():
            assert v == expected[k]
            assert isinstance(v, type(expected[k]))
        len(claims) == len(expected)

    @pytest.mark.parametrize(
        "json",
        [
            {
                "iss": "coap://as.example.com",
                "sub": "erikw",
                "aud": "coap://light.example.com",
                "cti": "123",
                "exp": 1444064944,
                "nbf": 1443944944,
                "iat": 1443944944,
                "cnf": {
                    "jwk": {
                        "kty": "OKP",
                        "use": "sig",
                        "crv": "Ed25519",
                        "kid": "01",
                        "x": "2E6dX83gqD_D0eAmqnaHe1TC1xuld6iAKXfw2OVATr0",
                        "alg": "EdDSA",
                    },
                },
            },
        ],
    )
    def test_claims_builder_from_json_with_cnf(self, ctx, json):
        claims = ctx.from_json(json)
        assert claims.iss == "coap://as.example.com"
        assert claims.sub == "erikw"
        assert claims.aud == "coap://light.example.com"
        assert claims.cti == "123"
        assert claims.exp == 1444064944
        assert claims.nbf == 1443944944
        assert claims.iat == 1443944944
        assert isinstance(claims.cnf, dict)

    def test_claims_builder_from_json_with_empty_object(self, ctx):
        claims = ctx.from_json({})
        assert claims.iss is None
        assert claims.sub is None
        assert claims.aud is None
        assert claims.cti is None
        assert claims.exp is None
        assert claims.nbf is None
        assert claims.iat is None
        assert claims.cnf is None

    @pytest.mark.parametrize(
        "invalid, msg",
        [
            (
                {1: "coap://as.example.com"},
                "It is already CBOR-like format.",
            ),
            (
                {"cnf": "xxx"},
                "cnf value should be dict.",
            ),
            (
                {"cnf": {"foo": "bar"}},
                "Supported cnf value not found.",
            ),
        ],
    )
    def test_claims_builder_from_json_with_invalid_arg(self, ctx, invalid, msg):
        with pytest.raises(ValueError) as err:
            ctx.from_json(invalid)
            pytest.fail("from_json should fail.")
        assert msg in str(err.value)

    def test_claims_builder_from_json_with_undefined_key(self, ctx):
        ctx.set_private_claim_names({"ext": -70001})
        claims = ctx.from_json(
            {
                "iss": "coap://as.example.com",
                "ext1": "foo",
            }
        )
        assert claims.get("ext1") is None

    def test_claims_builder_from_json_with_unknown_key(self, ctx):
        claims = ctx.from_json(
            {
                "iss": "coap://as.example.com",
                "unknown": "something",
            }
        ).to_dict()
        assert len(claims) == 1
        assert claims[1] == "coap://as.example.com"

    def test_claims_builder_set_private_claim_names(self, ctx):
        ctx.set_private_claim_names({"ext": -70001})
        claims = ctx.from_json(
            {
                "iss": "coap://as.example.com",
                "ext": "foo",
            }
        ).to_dict()
        assert len(claims) == 2
        assert claims[1] == "coap://as.example.com"
        assert claims[-70001] == "foo"

    def test_claims_builder_set_private_claim_names_with_invalid_key(self, ctx):
        with pytest.raises(ValueError) as err:
            ctx.set_private_claim_names({"ext": 1})
            pytest.fail("set_private_claim_names should fail.")
        assert (
            "The claim key should be other than the values listed in https://python-cwt.readthedocs.io/en/stable/claims.html."
            in str(err.value)
        )



# pylint: disable=R0201, R0904, W0621
# R0201: Method could be a function
# R0904: Too many public methods
# W0621: Redefined outer name

"""
Tests for COSE.
"""

import base64
import datetime

import cbor2
import pytest
from cbor2 import CBORTag

import cwt
from cwt import COSE, EncodeError, Recipient, cose_key

from .utils import key_path


@pytest.fixture(scope="session", autouse=True)
def ctx():
    return COSE(options={"kid_auto_inclusion": False})


class TestCOSE:
    """
    Tests for COSE.
    """

    def test_cose_constructor_with_options(self):
        ctx = COSE(options={"kid_auto_inclusion": False, "alg_auto_inclusion": False})
        assert isinstance(ctx, COSE)

    def test_cose_encode_and_decode_with_options(self):
        ctx = COSE(options={"kid_auto_inclusion": False, "alg_auto_inclusion": False})

        # MAC0
        mac_key = cose_key.from_symmetric_key(alg="HS256", kid="01")
        token = ctx.encode_and_mac(b"Hello world!", mac_key)
        assert b"Hello world!" == ctx.decode(token, mac_key)

        # MAC
        token = ctx.encode_and_mac(
            b"Hello world!",
            mac_key,
            recipients=[Recipient(unprotected={1: -6, 4: b"01"})],
        )
        assert b"Hello world!" == ctx.decode(token, mac_key)

        # Encrypt0
        enc_key = cose_key.from_symmetric_key(alg="ChaCha20/Poly1305", kid="02")
        token = ctx.encode_and_encrypt(b"Hello world!", enc_key)
        assert b"Hello world!" == ctx.decode(token, enc_key)

        # Encrypt
        token = ctx.encode_and_encrypt(
            b"Hello world!",
            enc_key,
            recipients=[Recipient(unprotected={1: -6, 4: b"02"})],
        )
        assert b"Hello world!" == ctx.decode(token, enc_key)

        # Signature1
        sig_key = cose_key.from_jwk(
            {
                "kty": "EC",
                "kid": "03",
                "crv": "P-256",
                "x": "usWxHK2PmfnHKwXPS54m0kTcGJ90UiglWiGahtagnv8",
                "y": "IBOL-C3BttVivg-lSreASjpkttcsz-1rb7btKLv8EX4",
                "d": "V8kgd2ZBRuh2dgyVINBUqpPDr7BOMGcF22CQMIUHtNM",
            }
        )
        token = ctx.encode_and_sign(b"Hello world!", sig_key)
        assert b"Hello world!" == ctx.decode(token, sig_key)

    def test_cose_encode_and_decode_with_protected_bytes(self):
        ctx = COSE(options={"kid_auto_inclusion": False, "alg_auto_inclusion": False})

        # MAC0
        mac_key = cose_key.from_symmetric_key(alg="HS256", kid="01")
        token = ctx.encode_and_mac(b"Hello world!", mac_key, protected=b"a0")
        assert b"Hello world!" == ctx.decode(token, mac_key)

        # MAC
        token = ctx.encode_and_mac(
            b"Hello world!",
            mac_key,
            protected=b"a0",
            recipients=[Recipient(unprotected={1: -6, 4: b"01"})],
        )
        assert b"Hello world!" == ctx.decode(token, mac_key)

        # Encrypt0
        enc_key = cose_key.from_symmetric_key(alg="ChaCha20/Poly1305", kid="02")
        token = ctx.encode_and_encrypt(b"Hello world!", enc_key, protected=b"a0")
        assert b"Hello world!" == ctx.decode(token, enc_key)

        # Encrypt
        token = ctx.encode_and_encrypt(
            b"Hello world!",
            enc_key,
            protected=b"a0",
            recipients=[Recipient(unprotected={1: -6, 4: b"02"})],
        )
        assert b"Hello world!" == ctx.decode(token, enc_key)

        # Signature1
        sig_key = cose_key.from_jwk(
            {
                "kty": "EC",
                "kid": "03",
                "crv": "P-256",
                "x": "usWxHK2PmfnHKwXPS54m0kTcGJ90UiglWiGahtagnv8",
                "y": "IBOL-C3BttVivg-lSreASjpkttcsz-1rb7btKLv8EX4",
                "d": "V8kgd2ZBRuh2dgyVINBUqpPDr7BOMGcF22CQMIUHtNM",
            }
        )
        token = ctx.encode_and_sign(b"Hello world!", sig_key, protected=b"a0")
        assert b"Hello world!" == ctx.decode(token, sig_key)

    def test_cose_constructor_with_invalid_kid_auto_inclusion(self):
        with pytest.raises(ValueError) as err:
            COSE(options={"kid_auto_inclusion": "xxx"})
            pytest.fail("COSE should fail.")
        assert "kid_auto_inclusion should be bool." in str(err.value)

    def test_cose_constructor_with_invalid_alg_auto_inclusion(self):
        with pytest.raises(ValueError) as err:
            COSE(options={"alg_auto_inclusion": "xxx"})
            pytest.fail("COSE should fail.")
        assert "alg_auto_inclusion should be bool." in str(err.value)

    def test_cose_sample_cose_wg_examples_hmac_01(self, ctx):
        cwt_str = "D8618543A10105A054546869732069732074686520636F6E74656E742E58202BDCC89F058216B8A208DDC6D8B54AA91F48BD63484986565105C9AD5A6682F6818340A20125044A6F75722D73656372657440"
        key = cose_key.from_jwk(
            {
                "kty": "oct",
                "alg": "HS256",
                "kid": "our-secret",
                "use": "sig",
                "k": "hJtXIZ2uSN5kbQfbtTNWbpdmhkV8FJG-Onbc6mxCcYg",
            }
        )
        token = ctx.encode_and_mac(
            b"This is the content.",
            key=key,
            recipients=[Recipient(unprotected={1: -6, 4: b"our-secret"})],
        )
        assert token == bytes.fromhex(cwt_str)
        assert ctx.decode(token, key) == b"This is the content."

    def test_cose_sample_cose_wg_examples_sign1_pass_01(self):
        key = cose_key.from_jwk(
            {
                "kty": "EC",
                "kid": "11",
                "crv": "P-256",
                "x": "usWxHK2PmfnHKwXPS54m0kTcGJ90UiglWiGahtagnv8",
                "y": "IBOL-C3BttVivg-lSreASjpkttcsz-1rb7btKLv8EX4",
                "d": "V8kgd2ZBRuh2dgyVINBUqpPDr7BOMGcF22CQMIUHtNM",
            }
        )
        ctx = COSE(options={"kid_auto_inclusion": False, "alg_auto_inclusion": False})
        token = ctx.encode_and_sign(
            b"This is the content.",
            key,
            protected=bytes.fromhex("a0"),
            unprotected={1: -7, 4: b"11"},
        )
        assert ctx.decode(token, key) == b"This is the content."

    def test_cose_sample_cose_wg_examples_sign_pass_01(self):
        key = cose_key.from_jwk(
            {
                "kty": "EC",
                "kid": "11",
                "crv": "P-256",
                "x": "usWxHK2PmfnHKwXPS54m0kTcGJ90UiglWiGahtagnv8",
                "y": "IBOL-C3BttVivg-lSreASjpkttcsz-1rb7btKLv8EX4",
                "d": "V8kgd2ZBRuh2dgyVINBUqpPDr7BOMGcF22CQMIUHtNM",
            }
        )
        ctx = COSE(options={"kid_auto_inclusion": False, "alg_auto_inclusion": False})
        token = ctx.encode_and_sign(
            b"This is the content.",
            [key],
            protected=bytes.fromhex("a0"),
        )
        assert ctx.decode(token, key) == b"This is the content."

    def test_cose_sample_cose_wg_examples_eddsa_01(self):
        cwt_str = "D8628443A10300A054546869732069732074686520636F6E74656E742E818343A10127A104423131584077F3EACD11852C4BF9CB1D72FABE6B26FBA1D76092B2B5B7EC83B83557652264E69690DBC1172DDC0BF88411C0D25A507FDB247A20C40D5E245FABD3FC9EC106"
        key = cose_key.from_jwk(
            {
                "kty": "OKP",
                "kid": "11",
                "crv": "Ed25519",
                "x": base64.urlsafe_b64encode(
                    bytes.fromhex(
                        "d75a980182b10ab7d54bfed3c964073a0ee172f3daa62325af021a68f707511a"
                    )
                )
                .replace(b"=", b"")
                .decode("ascii"),
                "d": base64.urlsafe_b64encode(
                    bytes.fromhex(
                        "9d61b19deffd5a60ba844af492ec2cc44449c5697b326919703bac031cae7f60"
                    )
                )
                .replace(b"=", b"")
                .decode("ascii"),
            }
        )
        ctx = COSE(options={"kid_auto_inclusion": False, "alg_auto_inclusion": False})
        token = ctx.encode_and_sign(
            b"This is the content.",
            [key],
            protected={3: 0},
        )
        assert token == bytes.fromhex(cwt_str)
        assert ctx.decode(token, key) == b"This is the content."

    def test_cose_sample_cose_wg_examples_eddsa_02(self):
        cwt_str = "D8628440A054546869732069732074686520636F6E74656E742E818343A10127A1044565643434385872ABF04F4BC7DFACF70C20C34A3CFBD27719911DC8518B2D67BF6AF62895D0FA1E6A1CB8B47AD1297C0E9C34BEB34E50DFFEF14350EBD57842807D54914111150F698543B0A5E1DA1DB79632C6415CE18EF74EDAEA680B0C8881439D869171481D78E2F7D26340C293C2ECDED8DE1425851900"
        key = cose_key.from_jwk(
            {
                "kty": "OKP",
                "kid": "ed448",
                "crv": "Ed448",
                "x": base64.urlsafe_b64encode(
                    bytes.fromhex(
                        "5fd7449b59b461fd2ce787ec616ad46a1da1342485a70e1f8a0ea75d80e96778edf124769b46c7061bd6783df1e50f6cd1fa1abeafe8256180"
                    )
                )
                .replace(b"=", b"")
                .decode("ascii"),
                "d": base64.urlsafe_b64encode(
                    bytes.fromhex(
                        "6c82a562cb808d10d632be89c8513ebf6c929f34ddfa8c9f63c9960ef6e348a3528c8a3fcc2f044e39a3fc5b94492f8f032e7549a20098f95b"
                    )
                )
                .replace(b"=", b"")
                .decode("ascii"),
            }
        )
        ctx = COSE(options={"kid_auto_inclusion": False, "alg_auto_inclusion": False})
        token = ctx.encode_and_sign(
            b"This is the content.",
            [key],
        )
        assert token == bytes.fromhex(cwt_str)
        assert ctx.decode(token, key) == b"This is the content."

    def test_cose_sample_cose_wg_examples_eddsa_sig_01(self):
        cwt_str = "D28445A201270300A10442313154546869732069732074686520636F6E74656E742E58407142FD2FF96D56DB85BEE905A76BA1D0B7321A95C8C4D3607C5781932B7AFB8711497DFA751BF40B58B3BCC32300B1487F3DB34085EEF013BF08F4A44D6FEF0D"
        key = cose_key.from_jwk(
            {
                "kty": "OKP",
                "kid": "11",
                "crv": "Ed25519",
                "x": base64.urlsafe_b64encode(
                    bytes.fromhex(
                        "d75a980182b10ab7d54bfed3c964073a0ee172f3daa62325af021a68f707511a"
                    )
                )
                .replace(b"=", b"")
                .decode("ascii"),
                "d": base64.urlsafe_b64encode(
                    bytes.fromhex(
                        "9d61b19deffd5a60ba844af492ec2cc44449c5697b326919703bac031cae7f60"
                    )
                )
                .replace(b"=", b"")
                .decode("ascii"),
            }
        )
        ctx = COSE(options={"alg_auto_inclusion": False})
        token = ctx.encode_and_sign(
            b"This is the content.",
            key,
            protected={1: -8, 3: 0},
        )
        assert token == bytes.fromhex(cwt_str)
        assert ctx.decode(token, key) == b"This is the content."

    def test_cose_sample_cose_wg_examples_eddsa_sig_02(self):
        cwt_str = "D28443A10127A10445656434343854546869732069732074686520636F6E74656E742E5872988240A3A2F189BD486DE14AA77F54686C576A09F2E7ED9BAE910DF9139C2AC3BE7C27B7E10A20FA17C9D57D3510A2CF1F634BC0345AB9BE00849842171D1E9E98B2674C0E38BFCF6C557A1692B01B71015A47AC9F7748840CAD1DA80CBB5B349309FEBB912672B377C8B2072AF1598B3700"
        key = cose_key.from_jwk(
            {
                "kty": "OKP",
                "kid": "ed448",
                "crv": "Ed448",
                "x": base64.urlsafe_b64encode(
                    bytes.fromhex(
                        "5fd7449b59b461fd2ce787ec616ad46a1da1342485a70e1f8a0ea75d80e96778edf124769b46c7061bd6783df1e50f6cd1fa1abeafe8256180"
                    )
                )
                .replace(b"=", b"")
                .decode("ascii"),
                "d": base64.urlsafe_b64encode(
                    bytes.fromhex(
                        "6c82a562cb808d10d632be89c8513ebf6c929f34ddfa8c9f63c9960ef6e348a3528c8a3fcc2f044e39a3fc5b94492f8f032e7549a20098f95b"
                    )
                )
                .replace(b"=", b"")
                .decode("ascii"),
            }
        )
        ctx = COSE()
        token = ctx.encode_and_sign(
            b"This is the content.",
            key,
        )
        assert token == bytes.fromhex(cwt_str)
        assert ctx.decode(token, key) == b"This is the content."

    def test_cose_sample_cose_wg_examples_aes_ccm_01(self, ctx):
        cwt_str = "D8608443A1010AA1054D89F52F65A1C580933B5261A72F581C6899DA0A132BD2D2B9B10915743EE1F7B92A46802388816C040275EE818340A20125044A6F75722D73656372657440"
        key = cose_key.from_jwk(
            {
                "kty": "oct",
                "alg": "AES-CCM-16-64-128",
                "kid": "our-secret",
                "use": "enc",
                "k": "hJtXIZ2uSN5kbQfbtTNWbg",
            }
        )
        token = ctx.encode_and_encrypt(
            b"This is the content.",
            key,
            nonce=bytes.fromhex("89F52F65A1C580933B5261A72F"),
            recipients=[Recipient(unprotected={1: -6, 4: b"our-secret"})],
        )
        assert token == bytes.fromhex(cwt_str)
        assert ctx.decode(token, key) == b"This is the content."

    def test_cose_sample_cose_wg_examples_aes_gcm_01(self, ctx):
        cwt_str = "D8608443A10101A1054C02D1F7E6F26C43D4868D87CE582460973A94BB2898009EE52ECFD9AB1DD25867374B3581F2C80039826350B97AE2300E42FC818340A20125044A6F75722D73656372657440"
        key = cose_key.from_jwk(
            {
                "kty": "oct",
                "alg": "A128GCM",
                "kid": "our-secret",
                "use": "enc",
                "k": "hJtXIZ2uSN5kbQfbtTNWbg",
            }
        )
        token = ctx.encode_and_encrypt(
            b"This is the content.",
            key,
            nonce=bytes.fromhex("02D1F7E6F26C43D4868D87CE"),
            recipients=[Recipient(unprotected={1: -6, 4: b"our-secret"})],
        )
        assert token == bytes.fromhex(cwt_str)
        assert ctx.decode(token, key) == b"This is the content."

    def test_cose_sample_cose_wg_examples_chacha_poly_01(self, ctx):
        # cwt_str = "D8608444A1011818A1054C26682306D4FB28CA01B43B8058245F2BD5381BBB04921A8477E55C0D850069674A05E683D416583AA0CEE0E2929CDF648094818340A2012504477365632D32353640"
        key = cose_key.from_jwk(
            {
                "kty": "oct",
                "alg": "ChaCha20/Poly1305",
                "kid": "sec-256",
                "use": "enc",
                "k": "Dx4tPEtaaXiHlqW0w9Lh8B8uPUxbanmIl6a1xNPi8QA",
            }
        )
        token = ctx.encode_and_encrypt(
            b"This is the content.",
            key,
            nonce=bytes.fromhex("26682306D4FB28CA01B43B80"),
            recipients=[Recipient(unprotected={1: -6, 4: b"sec-256"})],
        )
        # assert token == bytes.fromhex(cwt_str)
        assert ctx.decode(token, key) == b"This is the content."

    def test_cose_sample_cose_wg_examples_chacha_poly_enc_01(self, ctx):
        # cwt_str = "D08344A1011818A1054C5C3A9950BD2852F66E6C8D4F58243E536D4992A21591575C55FA22981B31AE1C045946D0E41A8A1ABD12BC9525922F4EB618"
        key = cose_key.from_jwk(
            {
                "kty": "oct",
                "alg": "ChaCha20/Poly1305",
                "kid": "sec-256",
                "use": "enc",
                "k": "Dx4tPEtaaXiHlqW0w9Lh8B8uPUxbanmIl6a1xNPi8QA",
            }
        )
        token = ctx.encode_and_encrypt(
            b"This is the content.",
            key,
            nonce=bytes.fromhex("5C3A9950BD2852F66E6C8D4F"),
        )
        # assert token == bytes.fromhex(cwt_str)
        assert ctx.decode(token, key) == b"This is the content."

    def test_cose_encode_and_mac_with_recipient_has_unsupported_alg(self, ctx):
        key = cose_key.from_symmetric_key(alg="HS256")
        with pytest.raises(NotImplementedError) as err:
            ctx.encode_and_mac(
                b"This is the content.",
                key,
                recipients=[Recipient(unprotected={1: 0, 4: b"our-secret"})],
            )
            pytest.fail("encode_and_mac should fail.")
        assert "Algorithms other than direct are not supported for recipients." in str(
            err.value
        )

    def test_cose_encode_and_encrypt_with_recipient_has_unsupported_alg(self, ctx):
        key = cose_key.from_jwk(
            {
                "kty": "oct",
                "alg": "AES-CCM-16-64-128",
                "kid": "our-secret",
                "use": "enc",
                "k": "hJtXIZ2uSN5kbQfbtTNWbg",
            }
        )
        with pytest.raises(NotImplementedError) as err:
            ctx.encode_and_encrypt(
                b"This is the content.",
                key,
                nonce=bytes.fromhex("89F52F65A1C580933B5261A72F"),
                recipients=[Recipient(unprotected={1: 0, 4: b"our-secret"})],
            )
            pytest.fail("encode_and_encrypt should fail.")
        assert "Algorithms other than direct are not supported for recipients." in str(
            err.value
        )

    def test_cose_encode_and_mac_with_invalid_payload(self, ctx):
        key = cose_key.from_symmetric_key(alg="HS256")
        with pytest.raises(EncodeError) as err:
            ctx.encode_and_mac(datetime.datetime.now(), key, {}, {})
            pytest.fail("encode_and_mac should fail.")
        assert "Failed to encode." in str(err.value)

    @pytest.mark.parametrize(
        "invalid, msg",
        [
            (
                "invalid_string_data",
                "Invalid COSE format.",
            ),
            (
                {},
                "Invalid COSE format.",
            ),
            (
                [],
                "Invalid COSE format.",
            ),
            (
                123,
                "Invalid COSE format.",
            ),
            (
                cbor2.CBORTag(16, "invalid_string_data"),
                "Invalid Encrypt0 format.",
            ),
            (
                cbor2.CBORTag(16, {}),
                "Invalid Encrypt0 format.",
            ),
            (
                cbor2.CBORTag(16, []),
                "Invalid Encrypt0 format.",
            ),
            (
                cbor2.CBORTag(16, 123),
                "Invalid Encrypt0 format.",
            ),
            (
                cbor2.CBORTag(16, [b"", b"invalid byte data", b""]),
                "unprotected header should be dict.",
            ),
            (
                cbor2.CBORTag(16, [b"", "invalid string data", b""]),
                "unprotected header should be dict.",
            ),
            (
                cbor2.CBORTag(16, [b"", [], b""]),
                "unprotected header should be dict.",
            ),
            (
                cbor2.CBORTag(16, [b"", 123, b""]),
                "unprotected header should be dict.",
            ),
            (
                cbor2.dumps(CBORTag(96, [])),
                "Invalid Encrypt format.",
            ),
            (
                cbor2.dumps(CBORTag(96, {})),
                "Invalid Encrypt format.",
            ),
            (
                cbor2.dumps(CBORTag(96, b"")),
                "Invalid Encrypt format.",
            ),
            (
                cbor2.dumps(CBORTag(96, 123)),
                "Invalid Encrypt format.",
            ),
            (
                cbor2.dumps(CBORTag(96, [b"", b"", b""])),
                "Invalid Encrypt format.",
            ),
            (
                cbor2.dumps(CBORTag(96, [b"", b"", b"", [b""]])),
                "unprotected header should be dict.",
            ),
            (
                cbor2.CBORTag(17, "invalid_string_data"),
                "Invalid MAC0 format.",
            ),
            (
                cbor2.CBORTag(17, {}),
                "Invalid MAC0 format.",
            ),
            (
                cbor2.CBORTag(17, []),
                "Invalid MAC0 format.",
            ),
            (
                cbor2.CBORTag(17, 123),
                "Invalid MAC0 format.",
            ),
            (
                cbor2.CBORTag(97, "invalid_string_data"),
                "Invalid MAC format.",
            ),
            (
                cbor2.CBORTag(97, {}),
                "Invalid MAC format.",
            ),
            (
                cbor2.CBORTag(97, []),
                "Invalid MAC format.",
            ),
            (
                cbor2.CBORTag(97, 123),
                "Invalid MAC format.",
            ),
            (
                cbor2.CBORTag(18, "invalid_string_data"),
                "Invalid Signature1 format.",
            ),
            (
                cbor2.CBORTag(18, {}),
                "Invalid Signature1 format.",
            ),
            (
                cbor2.CBORTag(18, []),
                "Invalid Signature1 format.",
            ),
            (
                cbor2.CBORTag(18, 123),
                "Invalid Signature1 format.",
            ),
            (
                cbor2.dumps(CBORTag(98, [])),
                "Invalid Signature format.",
            ),
            (
                cbor2.dumps(CBORTag(98, {})),
                "Invalid Signature format.",
            ),
            (
                cbor2.dumps(CBORTag(98, b"")),
                "Invalid Signature format.",
            ),
            (
                cbor2.dumps(CBORTag(98, 123)),
                "Invalid Signature format.",
            ),
            (
                cbor2.dumps(CBORTag(98, [b"", b"", b"", b""])),
                "Invalid Signature format.",
            ),
            (
                cbor2.dumps(CBORTag(98, [b"", b"", b"", [b""]])),
                "Invalid Signature format.",
            ),
        ],
    )
    def test_cose_decode_with_invalid_data(self, ctx, invalid, msg):
        with open(key_path("public_key_es256.pem")) as key_file:
            public_key = cose_key.from_pem(key_file.read(), kid="01")

        with pytest.raises(ValueError) as err:
            ctx.decode(invalid, public_key)
            pytest.fail("decode should fail.")
        assert msg in str(err.value)

    def test_cose_decode_mac0_with_invalid_multiple_keys(self, ctx):
        key1 = cose_key.from_symmetric_key(alg="HS256")
        key2 = cose_key.from_symmetric_key(alg="HS256")
        encoded = cwt.encode({"iss": "coap://as.example"}, key1)
        with pytest.raises(ValueError) as err:
            ctx.decode(encoded, [key1, key2])
            pytest.fail("decode should fail.")
        assert "key is not specified." in str(err.value)

    def test_cose_decode_encrypt0_with_invalid_multiple_keys(self, ctx):
        key1 = cose_key.from_symmetric_key(alg="ChaCha20/Poly1305")
        key2 = cose_key.from_symmetric_key(alg="ChaCha20/Poly1305")
        encoded = cwt.encode({"iss": "coap://as.example"}, key1)
        with pytest.raises(ValueError) as err:
            ctx.decode(encoded, [key1, key2])
            pytest.fail("decode should fail.")
        assert "key is not specified." in str(err.value)

    def test_cose_decode_signature1_with_invalid_multiple_keys(self, ctx):
        with open(key_path("public_key_es256.pem")) as key_file:
            key1 = cose_key.from_pem(key_file.read())
        with open(key_path("public_key_ed25519.pem")) as key_file:
            key2 = cose_key.from_pem(key_file.read())
        with open(key_path("private_key_ed25519.pem")) as key_file:
            private_key = cose_key.from_pem(key_file.read())
        encoded = cwt.encode({"iss": "coap://as.example"}, private_key)
        with pytest.raises(ValueError) as err:
            ctx.decode(encoded, [key1, key2])
            pytest.fail("decode should fail.")
        assert "key is not specified." in str(err.value)

    def test_cose_decode_with_key_not_found(self, ctx):
        with open(key_path("public_key_es256.pem")) as key_file:
            key1 = cose_key.from_pem(key_file.read(), kid="01")
        with open(key_path("public_key_ed25519.pem")) as key_file:
            key2 = cose_key.from_pem(key_file.read(), kid="02")
        with open(key_path("private_key_ed25519.pem")) as key_file:
            private_key = cose_key.from_pem(key_file.read(), kid="03")
        encoded = cwt.encode({"iss": "coap://as.example"}, private_key)
        with pytest.raises(ValueError) as err:
            ctx.decode(encoded, [key1, key2])
            pytest.fail("decode should fail.")
        assert "key is not specified." in str(err.value)



# pylint: disable=R0201, R0904, W0621
# R0201: Method could be a function
# R0904: Too many public methods
# W0621: Redefined outer name

"""
Tests for Claims.
"""
import pytest

from cwt.claims import Claims


class TestClaims:
    """
    Tests for Claims.
    """

    def test_claims_constructor(self):
        c = Claims({})
        assert isinstance(c, Claims)

    @pytest.mark.parametrize(
        "invalid, msg",
        [
            (
                {8: "xxx"},
                "cnf(8) should be dict.",
            ),
            (
                {8: {0: {}}},
                "cnf(8) should include COSE_Key, Encrypted_COSE_Key, or kid.",
            ),
            (
                {8: {1: "xxx"}},
                "COSE_Key in cnf(8) should be dict.",
            ),
            (
                {8: {2: "xxx"}},
                "Encrypted_COSE_Key in cnf(8) should be list.",
            ),
            (
                {8: {3: "xxx"}},
                "kid in cnf(8) should be bytes.",
            ),
        ],
    )
    def test_claims_constructor_with_invalid_arg(self, invalid, msg):
        with pytest.raises(ValueError) as err:
            Claims(invalid)
            pytest.fail("Claims should fail.")
        assert msg in str(err.value)



from typing import Any, Dict, List, Union

from .const import CWT_CLAIM_NAMES


class Claims:
    """
    A class for handling CWT Claims like JWT claims.
    """

    def __init__(
        self,
        claims: Dict[int, Any],
        claim_names: Dict[str, int] = CWT_CLAIM_NAMES,
    ):
        if -260 in claims and not isinstance(claims[-260], dict):
            raise ValueError("hcert(-260) should be map.")
        if -259 in claims and not isinstance(claims[-259], bytes):
            raise ValueError("EUPHNonce(-259) should be bstr.")
        if -258 in claims and not isinstance(claims[-258], bytes):
            raise ValueError("EATMAROEPrefix(-258) should be bstr.")
        if -257 in claims and not isinstance(claims[-257], list):
            raise ValueError("EAT-FDO(-257) should be array.")
        if 1 in claims and not isinstance(claims[1], str):
            raise ValueError("iss(1) should be str.")
        if 2 in claims and not isinstance(claims[2], str):
            raise ValueError("sub(2) should be str.")
        if 3 in claims:
            if not isinstance(claims[3], str) and not isinstance(claims[3], list):
                raise ValueError("aud(3) should be str or list[str].")
            if isinstance(claims[3], list):
                for c in claims[3]:
                    if not isinstance(c, str):
                        raise ValueError("aud(3) should be str or list[str].")
        if 4 in claims and not (
            isinstance(claims[4], int) or isinstance(claims[4], float)
        ):
            raise ValueError("exp(4) should be int or float.")
        if 5 in claims and not (
            isinstance(claims[5], int) or isinstance(claims[5], float)
        ):
            raise ValueError("nbf(5) should be int or float.")
        if 6 in claims and not (
            isinstance(claims[6], int) or isinstance(claims[6], float)
        ):
            raise ValueError("iat(6) should be int or float.")
        if 7 in claims and not isinstance(claims[7], bytes):
            raise ValueError("cti(7) should be bytes.")
        if 8 in claims:
            if not isinstance(claims[8], dict):
                raise ValueError("cnf(8) should be dict.")
            if 1 in claims[8]:
                if not isinstance(claims[8][1], dict):
                    raise ValueError("COSE_Key in cnf(8) should be dict.")
            elif 2 in claims[8]:
                if not isinstance(claims[8][2], list):
                    raise ValueError("Encrypted_COSE_Key in cnf(8) should be list.")
            elif 3 in claims[8]:
                if not isinstance(claims[8][3], bytes):
                    raise ValueError("kid in cnf(8) should be bytes.")
            else:
                raise ValueError(
                    "cnf(8) should include COSE_Key, Encrypted_COSE_Key, or kid."
                )
        self._claims = claims
        self._claim_names = claim_names
        return

    @property
    def iss(self) -> Union[str, None]:
        return self._claims.get(1, None)

    @property
    def sub(self) -> Union[str, None]:
        return self._claims.get(2, None)

    @property
    def aud(self) -> Union[str, None]:
        return self._claims.get(3, None)

    @property
    def exp(self) -> Union[int, None]:
        return self._claims.get(4, None)

    @property
    def nbf(self) -> Union[int, None]:
        return self._claims.get(5, None)

    @property
    def iat(self) -> Union[int, None]:
        return self._claims.get(6, None)

    @property
    def cti(self) -> Union[str, None]:
        if 7 not in self._claims:
            return None
        return self._claims[7].decode("utf-8")

    @property
    def cnf(self) -> Union[Dict[int, Any], List[Any], str, None]:
        if 8 not in self._claims:
            return None
        if 1 in self._claims[8]:
            key: Dict[int, Any] = self._claims[8][1]
            return key
        if 2 in self._claims[8]:
            eck: List[Any] = self._claims[8][2]
            return eck
        kid: bytes = self._claims[8][3]
        return kid.decode("utf-8")

    def get(self, key: Union[str, int]) -> Any:
        """
        Gets a claim value with a claim key.

        Args:
            key (Union[str, int]): A claim key.
        Returns:
            Any: The value of the claim.
        """
        int_key = 0
        if isinstance(key, str):
            int_key = self._claim_names.get(key, 0)
        else:
            int_key = key
        return self._claims.get(int_key, None) if int_key != 0 else None

    def to_dict(self) -> Dict[int, Any]:
        return self._claims



from typing import Any, Dict, List, Optional, Union

from cbor2 import CBORTag

from .cbor_processor import CBORProcessor
from .cose_key import COSEKey
from .recipient import Recipient, RecipientsBuilder


class COSE(CBORProcessor):
    """
    A COSE (CBOR Object Signing and Encryption) Implementaion built on top of
    `cbor2 <https://cbor2.readthedocs.io/en/stable/>`_.

    ``cwt.cose_key`` is a global object of this class initialized with default settings.
    """

    def __init__(self, options: Optional[Dict[str, Any]] = None):
        """
        Constructor.

        Args:
            options (Optional[Dict[str, Any]]): Options for the initial configuration
                of COSE. At this time, ``kid_auto_inclusion`` (default value: ``True``)
                and ``alg_auto_inclusion`` (default value: ``True``) are supported.
        """
        self._recipients_builder = RecipientsBuilder()
        self._kid_auto_inclusion = True
        self._alg_auto_inclusion = True
        if not options:
            return
        if "kid_auto_inclusion" in options:
            if not isinstance(options["kid_auto_inclusion"], bool):
                raise ValueError("kid_auto_inclusion should be bool.")
            self._kid_auto_inclusion = options["kid_auto_inclusion"]
        if "alg_auto_inclusion" in options:
            if not isinstance(options["alg_auto_inclusion"], bool):
                raise ValueError("alg_auto_inclusion should be bool.")
            self._alg_auto_inclusion = options["alg_auto_inclusion"]

    def encode_and_mac(
        self,
        payload: bytes,
        key: COSEKey,
        protected: Optional[Union[Dict[int, Any], bytes]] = None,
        unprotected: Optional[Dict[int, Any]] = None,
        recipients: Optional[List[Recipient]] = None,
        out: str = "",
    ) -> Union[bytes, CBORTag]:
        """
        Encodes data with MAC.

        Args:
            payload (bytes): A content to be MACed.
            key (COSEKey): A COSE key as a MAC Authentication key.
            protected (Union[Dict[int, Any], bytes]): Parameters that are to be cryptographically
                protected.
            unprotected (Dict[int, Any]): Parameters that are not cryptographically protected.
            recipients (Optional[List[Recipient]]): A list of recipient information structures.
            out(str): An output format. Only ``"cbor2/CBORTag"`` can be used. If ``"cbor2/CBORTag"``
                is specified. This function will return encoded data as
                `cbor2 <https://cbor2.readthedocs.io/en/stable/>`_'s ``CBORTag`` object.
                If any other value is specified, it will return encoded data as bytes.
        Returns:
            Union[bytes, CBORTag]: A byte string of the encoded COSE or a cbor2.CBORTag object.
        Raises:
            ValueError: Invalid arguments.
            EncodeError: Failed to encode data.
        """
        protected = {} if protected is None else protected
        unprotected = {} if unprotected is None else unprotected

        ctx = "MAC0" if not recipients else "MAC"
        b_protected = b""

        # MAC0
        if not recipients:
            if isinstance(protected, bytes):
                b_protected = protected
            else:
                if self._alg_auto_inclusion:
                    protected[1] = key.alg
                if self._kid_auto_inclusion and key.kid:
                    unprotected[4] = key.kid
                b_protected = self._dumps(protected)
            mac_structure = [ctx, b_protected, b"", payload]
            tag = key.sign(self._dumps(mac_structure))
            res = CBORTag(17, [b_protected, unprotected, payload, tag])
            return res if out == "cbor2/CBORTag" else self._dumps(res)

        # MAC
        recs = []
        for rec in recipients:
            recs.append(rec.to_list())
        if recipients[0].alg == -6:
            if not isinstance(protected, bytes):
                if self._alg_auto_inclusion:
                    protected[1] = key.alg
                if self._kid_auto_inclusion and key.kid:
                    unprotected[4] = key.kid
        else:
            raise NotImplementedError(
                "Algorithms other than direct are not supported for recipients."
            )

        if isinstance(protected, bytes):
            b_protected = protected
        else:
            b_protected = self._dumps(protected) if protected else b""
        mac_structure = [ctx, b_protected, b"", payload]
        tag = key.sign(self._dumps(mac_structure))
        cose_mac: List[Any] = [b_protected, unprotected, payload, tag]
        cose_mac.append(recs)
        res = CBORTag(97, cose_mac)
        return res if out == "cbor2/CBORTag" else self._dumps(res)

    def encode_and_sign(
        self,
        payload: bytes,
        key: Union[COSEKey, List[COSEKey]],
        protected: Optional[Union[Dict[int, Any], bytes]] = None,
        unprotected: Optional[Dict[int, Any]] = None,
        out: str = "",
    ) -> Union[bytes, CBORTag]:
        """
        Encodes data with signing.

        Args:
            payload (bytes): A content to be signed.
            key (Union[COSEKey, List[COSEKey]]): One or more COSE keys as signing keys.
            protected (Union[Dict[int, Any], bytes]): Parameters that are to be cryptographically
                protected.
            unprotected (Dict[int, Any]): Parameters that are not cryptographically
                protected.
            out(str): An output format. Only ``"cbor2/CBORTag"`` can be used. If ``"cbor2/CBORTag"``
                is specified. This function will return encoded data as
                `cbor2 <https://cbor2.readthedocs.io/en/stable/>`_'s ``CBORTag`` object.
                If any other value is specified, it will return encoded data as bytes.
        Returns:
            Union[bytes, CBORTag]: A byte string of the encoded COSE or a cbor2.CBORTag object.
        Raises:
            ValueError: Invalid arguments.
            EncodeError: Failed to encode data.
        """
        protected = {} if protected is None else protected
        unprotected = {} if unprotected is None else unprotected

        ctx = "Signature" if not isinstance(key, COSEKey) else "Signature1"
        if isinstance(key, COSEKey) and isinstance(protected, dict):
            if self._alg_auto_inclusion:
                protected[1] = key.alg
            if self._kid_auto_inclusion and key.kid:
                unprotected[4] = key.kid

        b_protected = b""
        if isinstance(protected, bytes):
            b_protected = protected
        else:
            b_protected = self._dumps(protected) if protected else b""

        # Signature1
        if isinstance(key, COSEKey):
            sig_structure = [ctx, b_protected, b"", payload]
            sig = key.sign(self._dumps(sig_structure))
            res = CBORTag(18, [b_protected, unprotected, payload, sig])
            return res if out == "cbor2/CBORTag" else self._dumps(res)

        # Signature
        sigs = []
        for k in key:
            p_header = self._dumps({1: k.alg})
            u_header = {4: k.kid} if k.kid else {}
            sig_structure = [ctx, b_protected, p_header, b"", payload]
            sig = k.sign(self._dumps(sig_structure))
            sigs.append([p_header, u_header, sig])
        res = CBORTag(98, [b_protected, unprotected, payload, sigs])
        return res if out == "cbor2/CBORTag" else self._dumps(res)

    def encode_and_encrypt(
        self,
        payload: bytes,
        key: COSEKey,
        protected: Optional[Union[Dict[int, Any], bytes]] = None,
        unprotected: Optional[Dict[int, Any]] = None,
        nonce: bytes = b"",
        recipients: Optional[List[Recipient]] = None,
        out: str = "",
    ) -> bytes:
        """
        Encodes data with encryption.

        Args:
            payload (bytes): A content to be encrypted.
            key (COSEKey): A COSE key as an encryption key.
            protected (Union[Dict[int, Any], bytes]): Parameters that are to be cryptographically
                protected.
            unprotected (Dict[int, Any]): Parameters that are not cryptographically
                protected.
            nonce (bytes): A nonce for encryption.
            recipients (Optional[List[Recipient]]): A list of recipient information structures.
            out(str): An output format. Only ``"cbor2/CBORTag"`` can be used. If ``"cbor2/CBORTag"``
                is specified. This function will return encoded data as
                `cbor2 <https://cbor2.readthedocs.io/en/stable/>`_'s ``CBORTag`` object.
                If any other value is specified, it will return encoded data as bytes.
        Returns:
            Union[bytes, CBORTag]: A byte string of the encoded COSE or a cbor2.CBORTag object.
        Raises:
            ValueError: Invalid arguments.
            EncodeError: Failed to encode data.
        """
        protected = {} if protected is None else protected
        unprotected = {} if unprotected is None else unprotected

        ctx = "Encrypt0" if not recipients else "Encrypt"

        if not nonce:
            try:
                nonce = key.generate_nonce()
            except NotImplementedError:
                raise ValueError(
                    "Nonce generation is not supported for the key. Set a nonce explicitly."
                )

        # Encrypt0
        if not recipients:
            if isinstance(protected, bytes):
                b_protected = protected
            else:
                if self._alg_auto_inclusion:
                    protected[1] = key.alg
                b_protected = self._dumps(protected) if protected else b""
            if self._kid_auto_inclusion and key.kid:
                unprotected[4] = key.kid
            unprotected[5] = nonce
            enc_structure = [ctx, b_protected, b""]
            aad = self._dumps(enc_structure)
            ciphertext = key.encrypt(payload, nonce, aad)
            res = CBORTag(16, [b_protected, unprotected, ciphertext])
            return res if out == "cbor2/CBORTag" else self._dumps(res)

        # Encrypt
        recs = []
        for rec in recipients:
            recs.append(rec.to_list())
        if recipients[0].alg == -6:
            if not isinstance(protected, bytes) and self._alg_auto_inclusion:
                protected[1] = key.alg
            if self._kid_auto_inclusion and key.kid:
                unprotected[4] = key.kid
            unprotected[5] = nonce
        else:
            raise NotImplementedError(
                "Algorithms other than direct are not supported for recipients."
            )
        if isinstance(protected, bytes):
            b_protected = protected
        else:
            b_protected = self._dumps(protected) if protected else b""
        enc_structure = [ctx, b_protected, b""]
        aad = self._dumps(enc_structure)
        ciphertext = key.encrypt(payload, nonce, aad)
        cose_enc: List[Any] = [b_protected, unprotected, ciphertext]
        cose_enc.append(recs)
        res = CBORTag(96, cose_enc)
        return res if out == "cbor2/CBORTag" else self._dumps(res)

    def decode(
        self, data: Union[bytes, CBORTag], key: Union[COSEKey, List[COSEKey]]
    ) -> bytes:
        """
        Verifies and decodes COSE data.

        Args:
            data (Union[bytes, CBORTag]): A byte string or cbor2.CBORTag of an
                encoded data.
            key (COSEKey): A COSE key to verify and decrypt the encoded data.
        Returns:
            bytes: A byte string of decoded payload.
        Raises:
            ValueError: Invalid arguments.
            DecodeError: Failed to decode data.
            VerifyError: Failed to verify data.
        """
        if isinstance(data, bytes):
            data = self._loads(data)
        if not isinstance(data, CBORTag):
            raise ValueError("Invalid COSE format.")

        keys: List[COSEKey] = key if isinstance(key, list) else [key]

        # Encrypt0
        if data.tag == 16:
            keys = self._filter_by_key_ops(keys, 4)
            if not isinstance(data.value, list) or len(data.value) != 3:
                raise ValueError("Invalid Encrypt0 format.")

            aad = self._dumps(["Encrypt0", data.value[0], b""])
            unprotected = data.value[1]
            if not isinstance(unprotected, dict):
                raise ValueError("unprotected header should be dict.")
            nonce = unprotected.get(5, None)
            k = self._get_key(keys, unprotected)
            if not k:
                raise ValueError("key is not specified.")
            return k.decrypt(data.value[2], nonce, aad)

        # Encrypt
        if data.tag == 96:
            keys = self._filter_by_key_ops(keys, 4)
            if not isinstance(data.value, list) or len(data.value) != 4:
                raise ValueError("Invalid Encrypt format.")

            aad = self._dumps(["Encrypt", data.value[0], b""])
            unprotected = data.value[1]
            if not isinstance(unprotected, dict):
                raise ValueError("unprotected header should be dict.")
            nonce = unprotected.get(5, None)
            recipients = self._recipients_builder.from_list(data.value[3])
            enc_key = recipients.derive_key(keys)
            return enc_key.decrypt(data.value[2], nonce, aad)

        # MAC0
        if data.tag == 17:
            keys = self._filter_by_key_ops(keys, 10)
            if not isinstance(data.value, list) or len(data.value) != 4:
                raise ValueError("Invalid MAC0 format.")

            msg = self._dumps(["MAC0", data.value[0], b"", data.value[2]])
            k = self._get_key(keys, data.value[1])
            if not k:
                raise ValueError("key is not specified.")
            k.verify(msg, data.value[3])
            return data.value[2]

        # MAC
        if data.tag == 97:
            keys = self._filter_by_key_ops(keys, 10)
            if not isinstance(data.value, list) or len(data.value) != 5:
                raise ValueError("Invalid MAC format.")
            to_be_maced = self._dumps(["MAC", data.value[0], b"", data.value[2]])
            recipients = self._recipients_builder.from_list(data.value[4])
            mac_auth_key = recipients.derive_key(keys)
            mac_auth_key.verify(to_be_maced, data.value[3])
            return data.value[2]

        # Signature1
        if data.tag == 18:
            keys = self._filter_by_key_ops(keys, 2)
            if not isinstance(data.value, list) or len(data.value) != 4:
                raise ValueError("Invalid Signature1 format.")

            to_be_signed = self._dumps(
                ["Signature1", data.value[0], b"", data.value[2]]
            )
            k = self._get_key(keys, data.value[1])
            if not k:
                raise ValueError("key is not specified.")
            k.verify(to_be_signed, data.value[3])
            return data.value[2]

        # Signature
        if data.tag == 98:
            keys = self._filter_by_key_ops(keys, 2)
            if not isinstance(data.value, list) or len(data.value) != 4:
                raise ValueError("Invalid Signature format.")
            sigs = data.value[3]
            if not isinstance(sigs, list):
                raise ValueError("Invalid Signature format.")
            for sig in sigs:
                if not isinstance(sig, list) or len(sig) != 3:
                    raise ValueError("Invalid Signature format.")
                k = self._get_key(keys, sig[1])
                if not k:
                    continue
                to_be_signed = self._dumps(
                    ["Signature", data.value[0], sig[0], b"", data.value[2]]
                )
                k.verify(to_be_signed, sig[2])
                return data.value[2]
            raise ValueError("Verification key not found.")
        raise ValueError(f"Unsupported or unknown CBOR tag({data.tag}).")

    def _get_key(
        self, keys: List[COSEKey], unprotected: Dict[int, Any]
    ) -> Union[COSEKey, None]:
        if len(keys) == 1:
            if 4 in unprotected and keys[0].kid:
                if unprotected[4] != keys[0].kid:
                    return None
            return keys[0]
        if 4 not in unprotected:
            return None
        for k in keys:
            if k.kid == unprotected[4]:
                return k
        return None

    def _filter_by_key_ops(self, keys: List[COSEKey], op: int) -> List[COSEKey]:
        res: List[COSEKey] = []
        for k in keys:
            if op in k.key_ops:
                res.append(k)
        if len(res) == 0:
            res = keys
        return res



# Registered CWT Claims
CWT_CLAIM_NAMES = {
    "hcert": -260,  # map
    "EUPHNonce": -259,  # bstr
    "EATMAROEPrefix": -258,  # bstr
    "EAT-FDO": -257,  # array
    "iss": 1,  # text string
    "sub": 2,  # text string
    "aud": 3,  # text string
    "exp": 4,  # integer or floating-point number
    "nbf": 5,  # integer or floating-point number
    "iat": 6,  # integer or floating-point number
    "cti": 7,  # byte string
    "cnf": 8,  # map
    "nonce": 10,  # bstr or list[bstr]
    "ueid": 11,  # bstr
    "oemid": 13,  # bstr
    "seclevel": 14,  # integer(1, 2, 3, 4)
    "secboot": 15,  # bool
    "dbgstat": 16,  # integer(0, 1, 2, 3, 4)
    "location": 17,  # map
    "eat_profile": 18,  # uri / oid
    "submods": 20,  # map
    # "origination": 30,  # tstr(string-or-uri)
    # "uptime": 31,  # uint
    # "chip-version": 32,
    # "board-version": 33,
    # "device-version": 34,
    # "chip-version-scheme": 35,
    # "board-version-scheme": 36,
    # "device-version-scheme": 37,
    # "ean-chip-version": 38,
    # "ean-board-version": 39,
    # "ean-device-version": 40,
    # "intuse": 41,  # integer(1, 2, 3, 4, 5)
}

# COSE key types
COSE_KEY_TYPES = {
    "OKP": 1,  # OCtet Key Pair
    "EC2": 2,  # Elliptic Curve Keys w/ x- and y-coordinate pair
    "EC": 2,  # Elliptic Curve Keys w/ x- and y-coordinate pair (JWK)
    "RSA": 3,  # RSA Key
    "Symmetric": 4,  # Symmetric Keys
    "oct": 4,  # Symmetric Keys (JWK)
    "HSS-LMS": 5,  # Public key for HSS/LMS hash-based digital signature
    "WalnutDSA": 6,  # WalnutDSA public key
}

# JWK Parameters
JWK_PARAMS_COMMON = {
    "kty": 1,
    "kid": 2,
    "alg": 3,
    # "use": *,
    "key_ops": 4,
}

JWK_PARAMS_OKP = {
    # "crv": -1,
    "x": -2,
    "d": -4,
}

JWK_PARAMS_EC = {
    # "crv": -1,
    "x": -2,
    "y": -3,
    "d": -4,
}

JWK_PARAMS_RSA = {
    "n": -1,
    "e": -2,
    "d": -3,
    "p": -4,
    "q": -5,
    "dp": -6,
    "dq": -7,
    "qi": -8,
    "oth": -9,
}

COSE_KEY_PARAMS_SYMMETRIC = {
    "k": -1,
}

# COSE key operation values.
COSE_KEY_OPERATION_VALUES = {
    "sign": 1,
    "verify": 2,
    "encrypt": 3,
    "decrypt": 4,
    "wrap key": 5,
    "unwrap key": 6,
    "derive key": 7,
    "derive bits": 8,
    "MAC create": 9,
    "MAC verify": 10,
    "wrapKey": 5,  # JWK
    "unwrapKey": 6,  # JWK
    "deriveKey": 7,  # JWK
    "deriveBits": 8,  # JWK
    "createMAC": 9,  # JWK-like lowerCamelCase
    "verifyMAC": 10,  # JWK-like lowerCamelCase
}

# COSE Algorithms for Content Encryption Key (CEK).
COSE_ALGORITHMS_CEK = {
    "A128GCM": 1,  # AES-GCM mode w/ 128-bit key, 128-bit tag
    "A192GCM": 2,  # AES-GCM mode w/ 192-bit key, 128-bit tag
    "A256GCM": 3,  # AES-GCM mode w/ 256-bit key, 128-bit tag
    "AES-CCM-16-64-128": 10,  # AES-CCM mode 128-bit key, 64-bit tag, 13-byte nonce
    "AES-CCM-16-64-256": 11,  # AES-CCM mode 256-bit key, 64-bit tag, 13-byte nonce
    "AES-CCM-64-64-128": 12,  # AES-CCM mode 128-bit key, 64-bit tag, 7-byte nonce
    "AES-CCM-64-64-256": 13,  # AES-CCM mode 256-bit key, 64-bit tag, 7-byte nonce
    "ChaCha20/Poly1305": 24,  # ChaCha20/Poly1305 w/ 256-bit key, 128-bit tag
    "AES-CCM-16-128-128": 30,  # AES-CCM mode 128-bit key, 128-bit tag, 13-byte nonce
    "AES-CCM-16-128-256": 31,  # AES-CCM mode 256-bit key, 128-bit tag, 13-byte nonce
    "AES-CCM-64-128-128": 32,  # AES-CCM mode 128-bit key, 128-bit tag, 7-byte nonce
    "AES-CCM-64-128-256": 33,  # AES-CCM mode 256-bit key, 128-bit tag, 7-byte nonce
    # etc.
}

# COSE Algorithms for MAC.
COSE_ALGORITHMS_MAC = {
    "HMAC 256/64": 4,  # HMAC w/ SHA-256 truncated to 64 bits
    "HS256": 5,  # HMAC w/ SHA-256 (JWK)
    "HS384": 6,  # HMAC w/ SHA-384 (JWK)
    "HS512": 7,  # HMAC w/ SHA-512 (JWK)
    "HMAC 256/256": 5,  # HMAC w/ SHA-256
    "HMAC 384/384": 6,  # HMAC w/ SHA-384
    "HMAC 512/512": 7,  # HMAC w/ SHA-512
    "AES-MAC128/64": 14,  # AES-MAC 128-bit key, 64-bit tag
    "AES-MAC256/64": 15,  # AES-MAC 256-bit key, 64-bit tag
    "AES-MAC128/128": 25,  # AES-MAC 128-bit key, 128-bit tag
    "AES-MAC256/128": 26,  # AES-MAC 256-bit key, 128-bit tag
    # etc.
}

# COSE Algorithms for Signature with RSA.
COSE_ALGORITHMS_SIG_RSA = {
    "R1": -65535,  # RSASSA-PKCS1-v1_5 using SHA-1 (No plan to support)
    "RS512": -259,  # RSASSA-PKCS1-v1_5 using SHA-512
    "RS384": -258,  # RSASSA-PKCS1-v1_5 using SHA-384
    "RS256": -257,  # RSASSA-PKCS1-v1_5 using SHA-256
    "PS512": -39,  # RSASSA-PSS w/ SHA-512
    "PS384": -38,  # RSASSA-PSS w/ SHA-384
    "PS256": -37,  # RSASSA-PSS w/ SHA-256
    # etc.
}

# JOSE Algorithms supported.
JOSE_ALGORITHMS_SUPPORTED = {
    "RS512": -259,
    "RS384": -258,
    "RS256": -257,
    "ES256K": -47,
    "PS512": -39,
    "PS384": -38,
    "PS256": -37,
    "ES512": -36,
    "ES384": -35,
    "EdDSA": -8,
    "ES256": -7,
    "dir": -6,
    "A128GCM": 1,
    "A192GCM": 2,
    "A256GCM": 3,
    "HS256": 5,
    "HS384": 6,
    "HS512": 7,
}

JWK_TYPES = {
    "OKP": 1,
    "EC": 2,  # EC2
    "RSA": 3,
    "oct": 4,  # Symmetric
}

JWK_OPERATIONS = {
    "sign": 1,
    "verify": 2,
    "encrypt": 3,
    "decrypt": 4,
    "wrapKey": 5,
    "unwrapKey": 6,
    "deriveKey": 7,
    "deriveBits": 8,
}

JWK_ELLIPTIC_CURVES = {
    "P-256": 1,
    "P-384": 2,
    "P-521": 3,
    "X25519": 4,
    "X448": 5,
    "Ed25519": 6,
    "Ed448": 7,
    "secp256k1": 8,
}

# COSE Algorithms for Symmetric Keys.
COSE_ALGORITHMS_SYMMETRIC = dict(COSE_ALGORITHMS_MAC, **COSE_ALGORITHMS_CEK)

# COSE Algorithms for RSA Keys.
COSE_ALGORITHMS_RSA = dict(COSE_ALGORITHMS_SIG_RSA)

# All of Supported COSE Algorithms.
COSE_ALGORITHMS = dict(COSE_ALGORITHMS_RSA, **COSE_ALGORITHMS_SYMMETRIC)

# COSE Named Algorithms for converting from JWK-like key.
COSE_NAMED_ALGORITHMS_SUPPORTED = dict(JOSE_ALGORITHMS_SUPPORTED, **COSE_ALGORITHMS)



"""
Tests for OKPKey.
"""
import pytest

from cwt.exceptions import VerifyError
from cwt.key_types.okp import OKPKey


class TestOKPKey:
    """
    Tests for OKPKey.
    """

    def test_okp_key_constructor_with_ed25519_key(self):
        private_key = OKPKey(
            {
                1: 1,
                3: -8,
                -1: 6,
                -2: b"\x18Es\xe0\x9a\x83\xfd\x0e\xe9K\xa8n\xf39i\x17\xfe\n2+|\xd1q\xcc\x87\xd2\xe9\xa9\xe8 \x9b\xd9",
                -4: b"B\xc6u\xd0|-\x07\xe7)\x8d\x1c\x13\x14\xa2\x8dFC1\xdf3sQ\x049|\x14\xc1\xed\x01\xe5\xdb\xa9",
            }
        )
        public_key = OKPKey(
            {
                1: 1,
                3: -8,
                -1: 6,
                -2: b"\x18Es\xe0\x9a\x83\xfd\x0e\xe9K\xa8n\xf39i\x17\xfe\n2+|\xd1q\xcc\x87\xd2\xe9\xa9\xe8 \x9b\xd9",
            }
        )
        assert private_key.kty == 1
        assert private_key.kid is None
        assert private_key.alg == -8
        assert private_key.crv == 6
        assert len(private_key.key_ops) == 2
        assert 1 in private_key.key_ops
        assert 2 in private_key.key_ops
        assert private_key.base_iv is None
        assert public_key.kty == 1
        assert public_key.kid is None
        assert public_key.alg == -8
        assert public_key.crv == 6
        assert len(public_key.key_ops) == 1
        assert 2 in public_key.key_ops
        assert public_key.base_iv is None
        try:
            sig = private_key.sign(b"Hello world!")
            public_key.verify(b"Hello world!", sig)
        except Exception:
            pytest.fail("sign/verify should not fail.")

    @pytest.mark.parametrize(
        "invalid, msg",
        [
            (
                {1: 2},
                "kty(1) should be OKP(1).",
            ),
            (
                {1: b"invalid"},
                "kty(1) should be int or str(tstr).",
            ),
            (
                {1: {}},
                "kty(1) should be int or str(tstr).",
            ),
            (
                {1: []},
                "kty(1) should be int or str(tstr).",
            ),
            (
                {1: 1},
                "x(-2) not found.",
            ),
            (
                {1: 1, -2: "xxxxxxxxxxxxxxxx"},
                "x(-2) should be bytes(bstr).",
            ),
            (
                {1: 1, -2: {}},
                "x(-2) should be bytes(bstr).",
            ),
            (
                {1: 1, -2: []},
                "x(-2) should be bytes(bstr).",
            ),
            (
                {
                    1: 1,
                    -2: b"\x18Es\xe0\x9a\x83\xfd\x0e\xe9K\xa8n\xf39i\x17\xfe\n2+|\xd1q\xcc\x87\xd2\xe9\xa9\xe8 \x9b\xd9",
                    3: -1,
                },
                "OKP algorithm mismatch: -1.",
            ),
            (
                {
                    1: 1,
                    -2: b"\x18Es\xe0\x9a\x83\xfd\x0e\xe9K\xa8n\xf39i\x17\xfe\n2+|\xd1q\xcc\x87\xd2\xe9\xa9\xe8 \x9b\xd9",
                },
                "crv(-1) not found.",
            ),
            (
                {
                    1: 1,
                    -2: b"\x18Es\xe0\x9a\x83\xfd\x0e\xe9K\xa8n\xf39i\x17\xfe\n2+|\xd1q\xcc\x87\xd2\xe9\xa9\xe8 \x9b\xd9",
                    -1: {},
                },
                "crv(-1) should be int or str(tstr).",
            ),
            (
                {
                    1: 1,
                    -2: b"\x18Es\xe0\x9a\x83\xfd\x0e\xe9K\xa8n\xf39i\x17\xfe\n2+|\xd1q\xcc\x87\xd2\xe9\xa9\xe8 \x9b\xd9",
                    -1: [],
                },
                "crv(-1) should be int or str(tstr).",
            ),
            (
                {
                    1: 1,
                    -2: b"\x18Es\xe0\x9a\x83\xfd\x0e\xe9K\xa8n\xf39i\x17\xfe\n2+|\xd1q\xcc\x87\xd2\xe9\xa9\xe8 \x9b\xd9",
                    -1: b"EdDSA",
                },
                "crv(-1) should be int or str(tstr).",
            ),
            (
                {
                    1: 1,
                    -2: b"\x18Es\xe0\x9a\x83\xfd\x0e\xe9K\xa8n\xf39i\x17\xfe\n2+|\xd1q\xcc\x87\xd2\xe9\xa9\xe8 \x9b\xd9",
                    -1: 3,
                },
                "Unsupported or unknown curve(3) for OKP.",
            ),
            (
                {
                    1: 1,
                    -2: b"\x18Es\xe0\x9a\x83\xfd\x0e\xe9K\xa8n\xf39i\x17\xfe\n2+|\xd1q\xcc\x87\xd2\xe9\xa9\xe8 \x9b\xd9",
                    -1: 8,
                },
                "Unsupported or unknown curve(8) for OKP.",
            ),
            (
                {1: 1, -2: b"invalid-x", -1: 6},
                "Invalid key parameter.",
            ),
            (
                {
                    1: 1,
                    -2: b"\x18Es\xe0\x9a\x83\xfd\x0e\xe9K\xa8n\xf39i\x17\xfe\n2+|\xd1q\xcc\x87\xd2\xe9\xa9\xe8 \x9b\xd9",
                    -4: "invalid-type-d",
                    -1: 6,
                },
                "d(-4) should be bytes(bstr).",
            ),
            (
                {
                    1: 1,
                    -2: b"\x18Es\xe0\x9a\x83\xfd\x0e\xe9K\xa8n\xf39i\x17\xfe\n2+|\xd1q\xcc\x87\xd2\xe9\xa9\xe8 \x9b\xd9",
                    -4: {},
                    -1: 6,
                },
                "d(-4) should be bytes(bstr).",
            ),
            (
                {
                    1: 1,
                    -2: b"\x18Es\xe0\x9a\x83\xfd\x0e\xe9K\xa8n\xf39i\x17\xfe\n2+|\xd1q\xcc\x87\xd2\xe9\xa9\xe8 \x9b\xd9",
                    -4: [],
                    -1: 6,
                },
                "d(-4) should be bytes(bstr).",
            ),
            (
                {
                    1: 1,
                    -2: b"\x18Es\xe0\x9a\x83\xfd\x0e\xe9K\xa8n\xf39i\x17\xfe\n2+|\xd1q\xcc\x87\xd2\xe9\xa9\xe8 \x9b\xd9",
                    -4: 123,
                    -1: 6,
                },
                "d(-4) should be bytes(bstr).",
            ),
            (
                {
                    1: 1,
                    -2: b"\x18Es\xe0\x9a\x83\xfd\x0e\xe9K\xa8n\xf39i\x17\xfe\n2+|\xd1q\xcc\x87\xd2\xe9\xa9\xe8 \x9b\xd9",
                    -4: b"invalid-d",
                    -1: 6,
                },
                "Invalid key parameter.",
            ),
        ],
    )
    def test_okp_key_constructor_with_invalid_args(self, invalid, msg):
        with pytest.raises(ValueError) as err:
            OKPKey(invalid)
            pytest.fail("OKPKey should fail.")
        assert msg in str(err.value)

    def test_okp_key_sign_with_es256_public_key(self):
        public_key = OKPKey(
            {
                1: 1,
                3: -8,
                -1: 6,
                -2: b"\x18Es\xe0\x9a\x83\xfd\x0e\xe9K\xa8n\xf39i\x17\xfe\n2+|\xd1q\xcc\x87\xd2\xe9\xa9\xe8 \x9b\xd9",
            }
        )
        with pytest.raises(ValueError) as err:
            public_key.sign(b"Hello world!")
            pytest.fail("sign should not fail.")
        assert "Public key cannot be used for signing." in str(err.value)

    def test_okp_key_verify_with_invalid_signature(self):
        private_key = OKPKey(
            {
                1: 1,
                3: -8,
                -1: 6,
                -2: b"\x18Es\xe0\x9a\x83\xfd\x0e\xe9K\xa8n\xf39i\x17\xfe\n2+|\xd1q\xcc\x87\xd2\xe9\xa9\xe8 \x9b\xd9",
                -4: b"B\xc6u\xd0|-\x07\xe7)\x8d\x1c\x13\x14\xa2\x8dFC1\xdf3sQ\x049|\x14\xc1\xed\x01\xe5\xdb\xa9",
            }
        )
        sig = private_key.sign(b"Hello world!")
        with pytest.raises(VerifyError) as err:
            private_key.verify(b"Hello world!", sig + b"xxx")
            pytest.fail("verify should not fail.")
        assert "Failed to verify." in str(err.value)



"""
Tests for EC2Key.
"""
import pytest

from cwt.exceptions import VerifyError
from cwt.key_types.ec2 import EC2Key


class TestEC2Key:
    """
    Tests for EC2Key.
    """

    def test_ec2_key_constructor_with_es256_key(self):
        private_key = EC2Key(
            {
                1: 2,
                3: -7,
                -2: b"\xa7\xddc*\xff\xc2?\x8b\xf8\x9c:\xad\xccDF\x9cZ \x04P\xef\x99\x0c=\xe6 w1\x08&\xba\xd9",
                -3: b"\xe2\xdb\xef\xfe\xb8\x8a\x12\xf27\xcb\x15:\x8a\xb9\x1a90B\x1a\x19^\xbc\xdc\xde\r\xb9s\xc1P\xf3\xaa\xdd",
                -4: b'\xe9\x16\x0c\xa96\x8d\xfa\xbc\xd5\xda"ua\xec\xf7\x96\r\x15\xf7_\xf3rb{\xb1\xde;\x99\x88\xafNh',
                -1: 1,
            }
        )
        public_key = EC2Key(
            {
                1: 2,
                3: -7,
                -2: b"\xa7\xddc*\xff\xc2?\x8b\xf8\x9c:\xad\xccDF\x9cZ \x04P\xef\x99\x0c=\xe6 w1\x08&\xba\xd9",
                -3: b"\xe2\xdb\xef\xfe\xb8\x8a\x12\xf27\xcb\x15:\x8a\xb9\x1a90B\x1a\x19^\xbc\xdc\xde\r\xb9s\xc1P\xf3\xaa\xdd",
                -1: 1,
            }
        )
        assert private_key.kty == 2
        assert private_key.kid is None
        assert private_key.alg == -7
        assert private_key.crv == 1
        assert len(private_key.key_ops) == 2
        assert 1 in private_key.key_ops
        assert 2 in private_key.key_ops
        assert private_key.base_iv is None
        assert public_key.kty == 2
        assert public_key.kid is None
        assert public_key.alg == -7
        assert public_key.crv == 1
        assert len(public_key.key_ops) == 1
        assert 2 in public_key.key_ops
        assert public_key.base_iv is None
        try:
            sig = private_key.sign(b"Hello world!")
            public_key.verify(b"Hello world!", sig)
        except Exception:
            pytest.fail("sign/verify should not fail.")

    def test_ec2_key_constructor_with_es384_key(self):
        private_key = EC2Key(
            {
                1: 2,
                3: -35,
                -2: b"\xec\xe6\xd0\xc1-\xd4>%\xb6\x0f\x9d\xbf\xe2\x89qB\xd7\x8f\xba\xa4\xe0\x97\xd0\x91\xcd\xbb\x90\x92,\xaa\xd4\x10D\xc35\xfe\x89\xbfs\xae,&\x8d\xef\xfa\xb0\xc0Q",
                -3: b'\xc9F\xd4\xc8\x97\xd9G\xb80Z\x96E:\x89U/\x89|c\xb2\x9d\x1e\x0ep\xf1\xc4\xedl\x99K9.\x882\x06"\xb2\xa5\xdd\x17HW\x1f-r>Fg',
                -4: b"Q/x\xf6;9\xb7\xfb\x8d\xc3l\xc1\x1dx6Z4\xfa\x99=8Nj\x05g\xde\xb45\x00'VY\xab,\x92\x82{\x08\xd6\xe5\xff\xe3\xc4\xee\xacu>\x96",
                -1: 2,
            }
        )
        public_key = EC2Key(
            {
                1: 2,
                3: -35,
                -2: b"\xec\xe6\xd0\xc1-\xd4>%\xb6\x0f\x9d\xbf\xe2\x89qB\xd7\x8f\xba\xa4\xe0\x97\xd0\x91\xcd\xbb\x90\x92,\xaa\xd4\x10D\xc35\xfe\x89\xbfs\xae,&\x8d\xef\xfa\xb0\xc0Q",
                -3: b'\xc9F\xd4\xc8\x97\xd9G\xb80Z\x96E:\x89U/\x89|c\xb2\x9d\x1e\x0ep\xf1\xc4\xedl\x99K9.\x882\x06"\xb2\xa5\xdd\x17HW\x1f-r>Fg',
                -1: 2,
            }
        )
        assert private_key.kty == 2
        assert private_key.kid is None
        assert private_key.alg == -35
        assert private_key.crv == 2
        assert len(private_key.key_ops) == 2
        assert 1 in private_key.key_ops
        assert 2 in private_key.key_ops
        assert private_key.base_iv is None
        assert public_key.kty == 2
        assert public_key.kid is None
        assert public_key.alg == -35
        assert public_key.crv == 2
        assert len(public_key.key_ops) == 1
        assert 2 in public_key.key_ops
        assert public_key.base_iv is None
        try:
            sig = private_key.sign(b"Hello world!")
            public_key.verify(b"Hello world!", sig)
        except Exception:
            pytest.fail("sign/verify should not fail.")

    def test_ec2_key_constructor_with_es512_key(self):
        private_key = EC2Key(
            {
                1: 2,
                3: -36,
                -2: b"\x01iP\xcb*\xb4\x04\xa2\xf9d\x0f0{\n\x07>}|KZ\x81\xbd8\xb3N\x90\xb0\x10\xebk\xd2TBxR\xf6lNE\x92S\x80\xd0k|\xd9l\x044\xd8f\xee\xa6KQ\xcf\xa0\x01?g\x87\xcc\xb4\xd1\xce\x95",
                -3: b"\x00\r\xa0;\xcb\xae\x1f\x8e\xad\xc2\x82\xc8\x8e%\x94,\xd9\xe2t\xebG]\xc0\xb3I\xeec\xdf\xc5\x02\xd6c\xd9\xb4\xcd\xb8h\xc7l7\x07u`\xc3\x91\rl)\xb2\x07\x00\x10\xa07\xcd\x02N#\xac=L\x91~\xa2\xb26",
                -4: b'\x00V\xe5\x80\x13u\xc1\xb0\x8e\xf0\x98K\x0f\xc5\x14\xc55T\xb9\xbf\xd5o\xdc\xfa\x01\xf6\x91\xee\x85\x7fa,\x821\xdf\xdc\x17^\xd9G\x973V\xe9\xdd"s\xf4\x12\xd9:$\xbe\xc3\xad\xf7\x80"\x187\xc1\xa1\x9a\x1e@\xd2',
                -1: 3,
            }
        )
        public_key = EC2Key(
            {
                1: 2,
                3: -36,
                -2: b"\x01iP\xcb*\xb4\x04\xa2\xf9d\x0f0{\n\x07>}|KZ\x81\xbd8\xb3N\x90\xb0\x10\xebk\xd2TBxR\xf6lNE\x92S\x80\xd0k|\xd9l\x044\xd8f\xee\xa6KQ\xcf\xa0\x01?g\x87\xcc\xb4\xd1\xce\x95",
                -3: b"\x00\r\xa0;\xcb\xae\x1f\x8e\xad\xc2\x82\xc8\x8e%\x94,\xd9\xe2t\xebG]\xc0\xb3I\xeec\xdf\xc5\x02\xd6c\xd9\xb4\xcd\xb8h\xc7l7\x07u`\xc3\x91\rl)\xb2\x07\x00\x10\xa07\xcd\x02N#\xac=L\x91~\xa2\xb26",
                -1: 3,
            }
        )
        assert private_key.kty == 2
        assert private_key.kid is None
        assert private_key.alg == -36
        assert private_key.crv == 3
        assert len(private_key.key_ops) == 2
        assert 1 in private_key.key_ops
        assert 2 in private_key.key_ops
        assert private_key.base_iv is None
        assert public_key.kty == 2
        assert public_key.kid is None
        assert public_key.alg == -36
        assert public_key.crv == 3
        assert len(public_key.key_ops) == 1
        assert 2 in public_key.key_ops
        assert public_key.base_iv is None
        try:
            sig = private_key.sign(b"Hello world!")
            public_key.verify(b"Hello world!", sig)
        except Exception:
            pytest.fail("sign/verify should not fail.")

    def test_cose_key_constructor_without_cose_key(self):
        with pytest.raises(TypeError):
            EC2Key()
            pytest.fail("EC2Key should fail.")

    @pytest.mark.parametrize(
        "invalid, msg",
        [
            (
                {},
                "kty(1) not found.",
            ),
            (
                {1: 1},
                "kty(1) should be EC2(2).",
            ),
            (
                {1: b"invalid"},
                "kty(1) should be int or str(tstr).",
            ),
            (
                {1: {}},
                "kty(1) should be int or str(tstr).",
            ),
            (
                {1: []},
                "kty(1) should be int or str(tstr).",
            ),
            (
                {1: 2},
                "x(-2) not found.",
            ),
            (
                {1: 2, -2: "xxxxxxxxxxxxxxxx"},
                "x(-2) should be bytes(bstr).",
            ),
            (
                {1: 2, -2: {}},
                "x(-2) should be bytes(bstr).",
            ),
            (
                {1: 2, -2: []},
                "x(-2) should be bytes(bstr).",
            ),
            (
                {
                    1: 2,
                    -2: b"\xa7\xddc*\xff\xc2?\x8b\xf8\x9c:\xad\xccDF\x9cZ \x04P\xef\x99\x0c=\xe6 w1\x08&\xba\xd9",
                },
                "y(-3) not found.",
            ),
            (
                {
                    1: 2,
                    -2: b"\xa7\xddc*\xff\xc2?\x8b\xf8\x9c:\xad\xccDF\x9cZ \x04P\xef\x99\x0c=\xe6 w1\x08&\xba\xd9",
                    -3: "yyyyyyyyyyyyyyyy",
                },
                "y(-3) should be bytes(bstr).",
            ),
            (
                {
                    1: 2,
                    -2: b"\xa7\xddc*\xff\xc2?\x8b\xf8\x9c:\xad\xccDF\x9cZ \x04P\xef\x99\x0c=\xe6 w1\x08&\xba\xd9",
                    -3: {},
                },
                "y(-3) should be bytes(bstr).",
            ),
            (
                {
                    1: 2,
                    -2: b"\xa7\xddc*\xff\xc2?\x8b\xf8\x9c:\xad\xccDF\x9cZ \x04P\xef\x99\x0c=\xe6 w1\x08&\xba\xd9",
                    -3: [],
                },
                "y(-3) should be bytes(bstr).",
            ),
            (
                {
                    1: 2,
                    -2: b"\xa7\xddc*\xff\xc2?\x8b\xf8\x9c:\xad\xccDF\x9cZ \x04P\xef\x99\x0c=\xe6 w1\x08&\xba\xd9",
                    -3: b"\xe2\xdb\xef\xfe\xb8\x8a\x12\xf27\xcb\x15:\x8a\xb9\x1a90B\x1a\x19^\xbc\xdc\xde\r\xb9s\xc1P\xf3\xaa\xdd",
                },
                "crv(-1) not found.",
            ),
            (
                {
                    1: 2,
                    -2: b"\xa7\xddc*\xff\xc2?\x8b\xf8\x9c:\xad\xccDF\x9cZ \x04P\xef\x99\x0c=\xe6 w1\x08&\xba\xd9",
                    -3: b"\xe2\xdb\xef\xfe\xb8\x8a\x12\xf27\xcb\x15:\x8a\xb9\x1a90B\x1a\x19^\xbc\xdc\xde\r\xb9s\xc1P\xf3\xaa\xdd",
                    -1: 0,
                },
                "Unsupported or unknown crv: 0",
            ),
            (
                {
                    1: 2,
                    -2: b"\xa7\xddc*\xff\xc2?\x8b\xf8\x9c:\xad\xccDF\x9cZ \x04P\xef\x99\x0c=\xe6 w1\x08&\xba\xd9",
                    -3: b"\xe2\xdb\xef\xfe\xb8\x8a\x12\xf27\xcb\x15:\x8a\xb9\x1a90B\x1a\x19^\xbc\xdc\xde\r\xb9s\xc1P\xf3\xaa\xdd",
                    -1: {},
                },
                "crv(-1) should be int or str(tstr).",
            ),
            (
                {
                    1: 2,
                    -2: b"\xa7\xddc*\xff\xc2?\x8b\xf8\x9c:\xad\xccDF\x9cZ \x04P\xef\x99\x0c=\xe6 w1\x08&\xba\xd9",
                    -3: b"\xe2\xdb\xef\xfe\xb8\x8a\x12\xf27\xcb\x15:\x8a\xb9\x1a90B\x1a\x19^\xbc\xdc\xde\r\xb9s\xc1P\xf3\xaa\xdd",
                    -1: [],
                },
                "crv(-1) should be int or str(tstr).",
            ),
            (
                {
                    1: 2,
                    -2: b"\xa7\xddc*\xff\xc2?\x8b\xf8\x9c:\xad\xccDF\x9cZ \x04P\xef\x99\x0c=\xe6 w1\x08&\xba\xd9",
                    -3: b"\xe2\xdb\xef\xfe\xb8\x8a\x12\xf27\xcb\x15:\x8a\xb9\x1a90B\x1a\x19^\xbc\xdc\xde\r\xb9s\xc1P\xf3\xaa\xdd",
                    -1: b"P-256",
                },
                "crv(-1) should be int or str(tstr).",
            ),
            (
                {
                    1: 2,
                    -2: b"invalid-length-x",
                    -3: b"\xe2\xdb\xef\xfe\xb8\x8a\x12\xf27\xcb\x15:\x8a\xb9\x1a90B\x1a\x19^\xbc\xdc\xde\r\xb9s\xc1P\xf3\xaa\xdd",
                    -1: 1,
                    3: -8,
                },
                "EC2 algorithm mismatch: -8.",
            ),
            (
                {
                    1: 2,
                    -2: b"invalid-length-x",
                    -3: b"\xe2\xdb\xef\xfe\xb8\x8a\x12\xf27\xcb\x15:\x8a\xb9\x1a90B\x1a\x19^\xbc\xdc\xde\r\xb9s\xc1P\xf3\xaa\xdd",
                    -1: 1,
                },
                "Coords should be 32 bytes for crv P-256.",
            ),
            (
                {
                    1: 2,
                    -2: b"invalid-length-x",
                    -3: b"\xe2\xdb\xef\xfe\xb8\x8a\x12\xf27\xcb\x15:\x8a\xb9\x1a90B\x1a\x19^\xbc\xdc\xde\r\xb9s\xc1P\xf3\xaa\xdd",
                    -1: 2,
                    3: -8,
                },
                "EC2 algorithm mismatch: -8.",
            ),
            (
                {
                    1: 2,
                    -2: b"invalid-length-x",
                    -3: b'\xc9F\xd4\xc8\x97\xd9G\xb80Z\x96E:\x89U/\x89|c\xb2\x9d\x1e\x0ep\xf1\xc4\xedl\x99K9.\x882\x06"\xb2\xa5\xdd\x17HW\x1f-r>Fg',
                    -1: 2,
                },
                "Coords should be 48 bytes for crv P-384.",
            ),
            (
                {
                    1: 2,
                    -2: b"invalid-length-x",
                    -3: b"\xe2\xdb\xef\xfe\xb8\x8a\x12\xf27\xcb\x15:\x8a\xb9\x1a90B\x1a\x19^\xbc\xdc\xde\r\xb9s\xc1P\xf3\xaa\xdd",
                    -1: 3,
                    3: -8,
                },
                "EC2 algorithm mismatch: -8.",
            ),
            (
                {
                    1: 2,
                    -2: b"invalid-length-x",
                    -3: b"\x00\r\xa0;\xcb\xae\x1f\x8e\xad\xc2\x82\xc8\x8e%\x94,\xd9\xe2t\xebG]\xc0\xb3I\xeec\xdf\xc5\x02\xd6c\xd9\xb4\xcd\xb8h\xc7l7\x07u`\xc3\x91\rl)\xb2\x07\x00\x10\xa07\xcd\x02N#\xac=L\x91~\xa2\xb26",
                    -1: 3,
                },
                "Coords should be 66 bytes for crv P-521.",
            ),
            (
                {
                    1: 2,
                    -2: b"invalid-length-x",
                    -3: b"\xe2\xdb\xef\xfe\xb8\x8a\x12\xf27\xcb\x15:\x8a\xb9\x1a90B\x1a\x19^\xbc\xdc\xde\r\xb9s\xc1P\xf3\xaa\xdd",
                    -1: 8,
                    3: -8,
                },
                "EC2 algorithm mismatch: -8.",
            ),
            (
                {
                    1: 2,
                    -2: b"invalid-length-x",
                    -3: b'6\x00\x14\xfd\x13\t\x07\xdc,\t\xda\x1c}G\x0f\xd5\x11\xb2H\xe8\xc9\x05\xa8\x1f\xf3Q?\xa3"\xec7A',
                    -1: 8,
                },
                "Coords should be 32 bytes for crv secp256k1.",
            ),
            (
                {
                    1: 2,
                    -2: b"\xa7\xddc*\xff\xc2?\x8b\xf8\x9c:\xad\xccDF\x9cZ \x04P\xef\x99\x0c=\xe6 w1\x08&\xba\xd9",
                    -3: b"\xe2\xdb\xef\xfe\xb8\x8a\x12\xf27\xcb\x15:\x8a\xb9\x1a90B\x1a\x19^\xbc\xdc\xde\r\xb9s\xc1P\xf3\xaa\xdd",
                    -4: "tstr-d",
                    -1: 1,
                },
                "d(-4) should be bytes(bstr).",
            ),
            (
                {
                    1: 2,
                    -2: b"\xa7\xddc*\xff\xc2?\x8b\xf8\x9c:\xad\xccDF\x9cZ \x04P\xef\x99\x0c=\xe6 w1\x08&\xba\xd9",
                    -3: b"\xe2\xdb\xef\xfe\xb8\x8a\x12\xf27\xcb\x15:\x8a\xb9\x1a90B\x1a\x19^\xbc\xdc\xde\r\xb9s\xc1P\xf3\xaa\xdd",
                    -4: {},
                    -1: 1,
                },
                "d(-4) should be bytes(bstr).",
            ),
            (
                {
                    1: 2,
                    -2: b"\xa7\xddc*\xff\xc2?\x8b\xf8\x9c:\xad\xccDF\x9cZ \x04P\xef\x99\x0c=\xe6 w1\x08&\xba\xd9",
                    -3: b"\xe2\xdb\xef\xfe\xb8\x8a\x12\xf27\xcb\x15:\x8a\xb9\x1a90B\x1a\x19^\xbc\xdc\xde\r\xb9s\xc1P\xf3\xaa\xdd",
                    -4: [],
                    -1: 1,
                },
                "d(-4) should be bytes(bstr).",
            ),
            (
                {
                    1: 2,
                    -2: b"\xa7\xddc*\xff\xc2?\x8b\xf8\x9c:\xad\xccDF\x9cZ \x04P\xef\x99\x0c=\xe6 w1\x08&\xba\xd9",
                    -3: b"\xe2\xdb\xef\xfe\xb8\x8a\x12\xf27\xcb\x15:\x8a\xb9\x1a90B\x1a\x19^\xbc\xdc\xde\r\xb9s\xc1P\xf3\xaa\xdd",
                    -4: b"invalid-length-d",
                    -1: 1,
                },
                "d(-4) should be 32 bytes for curve 1",
            ),
            (
                {
                    1: 2,
                    -2: b"\xa7\xddc*\xff\xc2?\x8b\xf8\x9c:\xad\xccDF\x9cZ \x04P\xef\x99\x0c=\xe6 w1\x08&\xba\xd9",
                    -3: b"\xe2\xdb\xef\xfe\xb8\x8a\x12\xf27\xcb\x15:\x8a\xb9\x1a90B\x1a\x19^\xbc\xdc\xde\r\xb9s\xc1P\xf3\xaa\xdd",
                    -4: b"dddddddddddddddddddddddddddddddd",
                    -1: 1,
                },
                "Invalid private key.",
            ),
        ],
    )
    def test_ec2_key_constructor_with_invalid_args(self, invalid, msg):
        with pytest.raises(ValueError) as err:
            EC2Key(invalid)
            pytest.fail("EC2Key should fail.")
        assert msg in str(err.value)

    def test_ec2_key_sign_with_es256_public_key(self):
        public_key = EC2Key(
            {
                1: 2,
                -2: b"\xa7\xddc*\xff\xc2?\x8b\xf8\x9c:\xad\xccDF\x9cZ \x04P\xef\x99\x0c=\xe6 w1\x08&\xba\xd9",
                -3: b"\xe2\xdb\xef\xfe\xb8\x8a\x12\xf27\xcb\x15:\x8a\xb9\x1a90B\x1a\x19^\xbc\xdc\xde\r\xb9s\xc1P\xf3\xaa\xdd",
                -1: 1,
            }
        )
        with pytest.raises(ValueError) as err:
            public_key.sign(b"Hello world!")
            pytest.fail("sign should not fail.")
        assert "Public key cannot be used for signing." in str(err.value)

    def test_ec2_key_verify_with_another_es256_public_key(self):
        private_key = EC2Key(
            {
                1: 2,
                -2: b"\xa7\xddc*\xff\xc2?\x8b\xf8\x9c:\xad\xccDF\x9cZ \x04P\xef\x99\x0c=\xe6 w1\x08&\xba\xd9",
                -3: b"\xe2\xdb\xef\xfe\xb8\x8a\x12\xf27\xcb\x15:\x8a\xb9\x1a90B\x1a\x19^\xbc\xdc\xde\r\xb9s\xc1P\xf3\xaa\xdd",
                -4: b'\xe9\x16\x0c\xa96\x8d\xfa\xbc\xd5\xda"ua\xec\xf7\x96\r\x15\xf7_\xf3rb{\xb1\xde;\x99\x88\xafNh',
                -1: 1,
            }
        )
        public_key2 = EC2Key(
            {
                1: 2,
                -1: 1,
                -2: b"\xd6\xdfNb\xfa-U\xab#\x85\xf4\xd1\xb5Z\x17m5WPN\xd12*\xe1\xc9\xdb_\xdcD!\xd1Y",
                -3: b"(\x8bw\x1d\xf5\xbe\x9a\xf3\x11\xbb\xe7\xa9\x98\xc9\xf9\\\xee\x862\x10\xae\x83\xb2\x97\xa9/\xd5\xb6~d\x85\x85",
                # -4: b'>\x89\xb8[2\xe3\xf1\xf9x\xea%\xa9\xaes\xf1\xb5\xfeo;\xfa\xb4\x13\xb33\x068j\x04\xfd\x03[\xe8'
            }
        )
        sig = private_key.sign(b"Hello world!")
        with pytest.raises(VerifyError) as err:
            public_key2.verify(b"Hello world!", sig)
            pytest.fail("verify should not fail.")
        assert "Failed to verify." in str(err.value)

    def test_ec2_key_verify_with_invalid_signature(self):
        private_key = EC2Key(
            {
                1: 2,
                -2: b"\xa7\xddc*\xff\xc2?\x8b\xf8\x9c:\xad\xccDF\x9cZ \x04P\xef\x99\x0c=\xe6 w1\x08&\xba\xd9",
                -3: b"\xe2\xdb\xef\xfe\xb8\x8a\x12\xf27\xcb\x15:\x8a\xb9\x1a90B\x1a\x19^\xbc\xdc\xde\r\xb9s\xc1P\xf3\xaa\xdd",
                -4: b'\xe9\x16\x0c\xa96\x8d\xfa\xbc\xd5\xda"ua\xec\xf7\x96\r\x15\xf7_\xf3rb{\xb1\xde;\x99\x88\xafNh',
                -1: 1,
            }
        )
        sig = private_key.sign(b"Hello world!")
        with pytest.raises(VerifyError) as err:
            private_key.verify(b"Hello world!", sig + b"xxx")
            pytest.fail("verify should not fail.")
        assert "Invalid signature." in str(err.value)



# pylint: disable=R0201, R0904, W0621
# R0201: Method could be a function
# R0904: Too many public methods
# W0621: Redefined outer name

"""
Tests for Recipient.
"""
import cbor2
import pytest

from cwt import Recipient, cose_key
from cwt.recipient import Recipients, RecipientsBuilder


@pytest.fixture(scope="session", autouse=True)
def ctx():
    return Recipient()


class TestRecipient:
    """
    Tests for Recipient.
    """

    def test_recipient_constructor(self):
        r = Recipient()
        assert isinstance(r, Recipient)
        assert r.protected == {}
        assert r.unprotected == {}
        assert r.ciphertext == b""
        assert isinstance(r.recipients, list)
        assert r.kid == b""
        assert r.alg == 0
        assert len(r.recipients) == 0
        res = r.to_list()
        assert len(res) == 3
        assert res[0] == b""
        assert isinstance(res[1], dict)
        assert res[2] == b""

    def test_recipient_constructor_with_args(self):
        child = Recipient(unprotected={1: -6, 4: b"our-secret"})
        r = Recipient(
            protected={"foo": "bar"},
            unprotected={1: -1, 4: b"our-secret"},
            recipients=[child],
        )
        assert isinstance(r.protected, dict)
        assert r.protected["foo"] == "bar"
        assert isinstance(r.unprotected, dict)
        assert r.kid == b"our-secret"
        assert r.alg == -1
        assert r.ciphertext == b""
        assert len(r.recipients) == 1
        res = r.to_list()
        assert len(res) == 4
        assert isinstance(res[3], list)
        assert len(res[3]) == 1
        assert isinstance(res[3][0], list)
        assert len(res[3][0]) == 3

    def test_recipient_constructor_with_protected_bytes(self):
        r = Recipient(protected=cbor2.dumps({"foo": "bar"}))
        assert isinstance(r.protected, dict)
        assert r.protected["foo"] == "bar"

    def test_recipient_constructor_with_empty_recipients(self):
        r = Recipient(unprotected={1: -6, 4: b"our-secret"}, recipients=[])
        assert isinstance(r, Recipient)
        assert r.protected == {}
        assert isinstance(r.unprotected, dict)
        assert r.ciphertext == b""
        assert len(r.recipients) == 0
        res = r.to_list()
        assert len(res) == 3

    def test_recipient_constructor_with_alg_a128kw(self):
        r = Recipient(unprotected={1: -3, 4: b"our-secret"})
        assert isinstance(r, Recipient)
        assert r.protected == {}
        assert isinstance(r.unprotected, dict)
        assert r.ciphertext == b""
        assert len(r.recipients) == 0
        res = r.to_list()
        assert len(res) == 3

    @pytest.mark.parametrize(
        "protected, unprotected, ciphertext, recipients, msg",
        [
            (
                {"foo": "bar"},
                {1: -6, 4: b"our-secret"},
                b"",
                [],
                "protected header should be empty.",
            ),
            (
                {},
                {1: -6, 4: b"our-secret"},
                b"xxx",
                [],
                "ciphertext should be zero-length bytes.",
            ),
            (
                {},
                {1: -6, 4: b"our-secret"},
                b"",
                [Recipient()],
                "recipients should be absent.",
            ),
        ],
    )
    def test_recipient_constructor_with_invalid_args(
        self, protected, unprotected, ciphertext, recipients, msg
    ):
        with pytest.raises(ValueError) as err:
            Recipient(protected, unprotected, ciphertext, recipients)
            pytest.fail("Recipient() should fail.")
        assert msg in str(err.value)

    def test_recipient_constructor_with_invalid_recipients(self):
        child = {}
        with pytest.raises(ValueError) as err:
            Recipient(unprotected={1: 0, 4: b"our-secret"}, recipients=[child])
            pytest.fail("Recipient() should fail.")
        assert "Invalid child recipient." in str(err.value)


class TestRecipients:
    """
    Tests for Recipients.
    """

    def test_recipients_constructor(self):
        r = Recipients([Recipient()])
        assert isinstance(r, Recipients)

    def test_recipients_constructor_with_recipient_alg_direct(self):
        key = cose_key.from_symmetric_key(
            "mysecret", alg="HMAC 256/64", kid="our-secret"
        )
        r = Recipients([Recipient(unprotected={1: -6, 4: b"our-secret"})])
        key = r.derive_key([key])
        assert key.kty == 4
        assert key.alg == 4
        assert key.kid == b"our-secret"

    def test_recipients_derive_key_with_empty_recipient(self):
        key = cose_key.from_symmetric_key(
            "mysecret", alg="HMAC 256/64", kid="our-secret"
        )
        r = Recipients([Recipient()])
        with pytest.raises(ValueError) as err:
            r.derive_key([key])
            pytest.fail("derive_key() should fail.")
        assert "Failed to derive a key." in str(err.value)

    def test_recipients_derive_key_without_key(self):
        r = Recipients([Recipient(unprotected={1: -6, 4: b"our-secret"})])
        with pytest.raises(ValueError) as err:
            r.derive_key([])
            pytest.fail("derive_key() should fail.")
        assert "Failed to derive a key." in str(err.value)

    def test_recipients_derive_key_with_different_kid(self):
        key = cose_key.from_symmetric_key(
            "mysecret", alg="HMAC 256/64", kid="our-secret"
        )
        r = Recipients([Recipient(unprotected={1: -6, 4: b"your-secret"})])
        with pytest.raises(ValueError) as err:
            r.derive_key([key])
            pytest.fail("derive_key() should fail.")
        assert "Failed to derive a key." in str(err.value)


class TestRecipientsBuilder:
    """
    Tests for RecipientsBuilder.
    """

    def test_recipients_builder_constructor(self):
        rb = RecipientsBuilder()
        assert isinstance(rb, RecipientsBuilder)

    def test_recipients_builder_from_list(self):
        rb = RecipientsBuilder()
        try:
            rb.from_list([[b"", {}, b""]])
        except Exception:
            pytest.fail("from_list() should not fail.")

    def test_recipients_builder_from_list_with_empty_recipients(self):
        rb = RecipientsBuilder()
        try:
            rb.from_list([[b"", {}, b"", []]])
        except Exception:
            pytest.fail("from_list() should not fail.")

    def test_recipients_builder_from_list_with_recipients(self):
        rb = RecipientsBuilder()
        try:
            rb.from_list([[b"", {}, b"", [[b"", {1: -6, 4: b"our-secret"}, b""]]]])
        except Exception:
            pytest.fail("from_list() should not fail.")

    @pytest.mark.parametrize(
        "invalid, msg",
        [
            ([{}], "Invalid recipient format."),
            ([123], "Invalid recipient format."),
            ([[]], "Invalid recipient format."),
            ([[b"", {}]], "Invalid recipient format."),
            ([[b"", {}, b"", [], {}]], "Invalid recipient format."),
            ([["", {}, b""]], "protected header should be bytes."),
            ([[{}, {}, b""]], "protected header should be bytes."),
            ([[[], {}, b""]], "protected header should be bytes."),
            ([[123, {}, b""]], "protected header should be bytes."),
            ([[b"", [], b""]], "unprotected header should be dict."),
            ([[b"", "", b""]], "unprotected header should be dict."),
            ([[b"", b"", b""]], "unprotected header should be dict."),
            ([[b"", 123, b""]], "unprotected header should be dict."),
            ([[b"", {}, ""]], "ciphertext should be bytes."),
            ([[b"", {}, {}]], "ciphertext should be bytes."),
            ([[b"", {}, []]], "ciphertext should be bytes."),
            ([[b"", {}, 123]], "ciphertext should be bytes."),
            ([[b"", {}, b"", {}]], "recipients should be list."),
            ([[b"", {}, b"", ""]], "recipients should be list."),
            ([[b"", {}, b"", b""]], "recipients should be list."),
            ([[b"", {}, b"", 123]], "recipients should be list."),
        ],
    )
    def test_recipients_builder_from_list_with_invalid_args(self, invalid, msg):
        rb = RecipientsBuilder()
        with pytest.raises(ValueError) as err:
            rb.from_list(invalid)
            pytest.fail("derive_key() should fail.")
        assert msg in str(err.value)
